'''
Defines classes to manage information for windows and the screen.

@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

from gi.repository import Gdk
from gi.repository import GdkPixbuf
import pyatspi

import xdg
import xdg.BaseDirectory
import xdg.DesktopEntry
import xdg.IconTheme

import datetime
import dbus
import json
import os
import subprocess
import sys


def get_window_manager():
  preferred = os.getenv('ACCERCISER_WINDOW_MANAGER')
  if preferred == 'kwin':
    return KWinWindowManager()
  if preferred == 'gnomeshell':
    return GnomeShellWindowManager()
  return WnckWindowManager()


class WindowGeometry:
  '''
  Class that represents the geometry (position and size) of a
  (system) window.
  '''

  def __init__(self, x, y, width, height):
    self.x = x
    self.y = y
    self.width = width
    self.height = height


class WindowInfo:
  '''
  Class that represents relevant information of a (system) window.
  '''

  def __init__(self, title, x, y, width, height, stacking_index=0,
               on_current_workspace=True, buffer_geometry=None,
               app_id=None, process_id=0):
    self.title = title
    self.x = x
    self.y = y
    self.width = width
    self.height = height
    self.stacking_index = stacking_index
    self.on_current_workspace = on_current_workspace
    self.buffer_geometry = buffer_geometry
    self.app_id = app_id
    self.process_id = process_id


class WindowManager:
  '''
  Class that provides information related to windows on the screen and the
  screen itself.
  This includes fetching information from the X11 window manager or
  Wayland compositor.
  Any handling specific to a particular window system should be done in
  (a subclass of) this class.
  '''

  def getToolkitNameAndVersion(self, acc):
    '''
    Return the name and major version number of the toolkit
    used by the application that the given accessible belongs to.

    @param acc: Accessible for whose application to retrieve
                the toolkit info.
    @type acc: Atspi.Accessible
    @return: The toolkit name and major version, or `(None, None)`
             if that information couldn't be retrieved.
    @rtype: tuple(str, int)
    '''
    app = acc.get_application()
    if app and app.role == pyatspi.ROLE_APPLICATION:
      toolkit = app.get_toolkit_name()
      version = app.get_toolkit_version()
      if not version or (not isinstance(version, str)):
        return True
      try:
        major_version = int(version.split('.')[0])
        return toolkit, major_version
      except ValueError:
        pass
    return None, None


  def supportsScreenCoords(self, acc):
    '''
    Returns False when the accessible does not support
    querying screen coordinates directly via AT-SPI,
    otherwise True.
    '''
    app = acc.get_application()
    toolkit, major_version = self.getToolkitNameAndVersion(acc)
    if toolkit and major_version:
      # Gtk 4 doesn't support global/screen coords
      if toolkit.lower() == 'gtk' and (major_version >= 4):
          return False
    return True

  def getWindowInfos(self):
    '''
    Retrieve window information for all (system) windows.

    The default implementation returns an empty list.
    Subclasses should override this.

    @return: Window information for all windows.
    @rtype: list[WindowInfo]
    '''
    return []

  def getWindowInfo(self, toplevel):
    '''
    Get information on the (system) window that the toplevel
    corresponds to, if possible.

    @param toplevel: The top level for which to receive the corresponding
                      window info.
    @type toplevel: Atspi.Accessible
    @return: The WindowInfo for the toplevel's system window, or None.
    @rtype: WindowInfo
    '''
    candidates = []
    win_infos = self.getWindowInfos()
    for window in win_infos:
      # for GTK 3 apps (at least seen with gedit and Evince), the window-local
      # coordinates for accessibles and size of the top-level frame refer to the
      # buffer geometry, so use that if available
      if window.buffer_geometry:
        toolkit, major_version = self.getToolkitNameAndVersion(toplevel)
        if toolkit and toolkit.lower() == 'gtk' and major_version == 3:
          window.x = window.buffer_geometry.x
          window.y = window.buffer_geometry.y
          window.width = window.buffer_geometry.width
          window.height = window.buffer_geometry.height

      # match by name
      if toplevel.name == window.title:
        candidates.append(window)

    window = None
    if len(candidates) == 1:
      window = candidates[0]
    elif len(candidates) > 1:
      # filter by process ID if any window has the same one
      # (not guaranteed e.g. for Flatpak)
      pid_candidates = []
      for candidate in candidates:
        if candidate.process_id == toplevel.get_process_id():
          pid_candidates.append(candidate)
      if pid_candidates:
        candidates = pid_candidates

      # in case of multiple candidates, prefer one for which size reported by AT-SPI matches
      atspi_width, atspi_height = toplevel.queryComponent().getSize()
      for candidate in candidates:
        if candidate.width == atspi_width and candidate.height == atspi_height:
          window = candidate
          break
      # if size doesn't match for any, use first candidate
      if window is None:
        window = candidates[0]

    return window

  def getCurrentWorkspaceWindowOrder(self):
    '''
    Get list of names of windows on the current workspace
    in stacking order.

    The list is in bottom-to-top order.
    '''
    windows = self.getWindowInfos()

    # filter out windows not on the current workspace
    windows = [win for win in windows if win.on_current_workspace]

    # sort according to stacking order
    def get_stacking_index(win):
      return win.stacking_index
    windows.sort(key=get_stacking_index)
    window_titles = [win.title for win in windows]
    return window_titles

  def getApplicationIcon(self, app):
    '''
    Get the icon of one of the top-level windows of the given application.

    The default implementation returns None.
    Subclasses should override this.

    @param app: The application accessible for which to retrieve a window icon.
    @type toplevel: Atspi.Accessible
    @return: The icon, or None.
    @rtype: GdkPixbuf.Pixbuf
    '''
    return None

  def getApplicationIconViaAppID(self, app):
    '''
    Get the icon of one of the top-level windows of the given application,
    using the app ID set for the top-level windows.

    This is primarily meant as a helper method to implement the
    above getApplicationIcon method in deriving classes.

    @param app: The application accessible for which to retrieve a window icon.
    @type toplevel: Atspi.Accessible
    @return: The icon, or None.
    @rtype: GdkPixbuf.Pixbuf
    '''
    # see also the FreeDesktop Desktop entry spec:
    # https://specifications.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html
    try:
      window_infos = self.getWindowInfos()
      for toplevel in app:
        for win in window_infos:
          if win.title == toplevel.name:
            desktop_file_name = win.app_id
            if desktop_file_name:
              for dir in xdg.BaseDirectory.xdg_data_dirs:
                desktop_file = os.path.join(dir, 'applications', desktop_file_name + '.desktop')
                if os.path.exists(desktop_file):
                  desktop_entry = xdg.DesktopEntry.DesktopEntry(desktop_file)
                  icon = desktop_entry.getIcon()
                  if icon:
                    icon_path = xdg.IconTheme.getIconPath(icon)
                    if icon_path:
                      # load pixbuf with max size of 24x24 pixel
                      pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(icon_path, 24, 24, True)
                      if pixbuf:
                        return pixbuf
    except Exception:
      pass

  def getScreenExtents(self, acc):
    '''
    Returns the extents of the given accessible object
    in screen/global coordinates.
    '''
    try:
      component_iface = acc.queryComponent()
    except NotImplementedError:
      return None

    screen_coords_supported = self.supportsScreenCoords(acc)

    if not screen_coords_supported:
      toplevel = acc
      while toplevel.parent and toplevel.parent.role != pyatspi.ROLE_APPLICATION:
        toplevel = toplevel.parent
      # try to get position info for the corresponding system window
      # and calculate screen coordinates from screen coords of the system
      # window and window-relative coords of the object
      win_info = self.getWindowInfo(toplevel)
      if win_info:
        extents = component_iface.getExtents(pyatspi.WINDOW_COORDS)
        extents.x += win_info.x
        extents.y += win_info.y
        return extents

    # query screen coords directly via AT-SPI
    extents = component_iface.getExtents(pyatspi.DESKTOP_COORDS)
    return extents

  def convertScreenToWindowCoords(self, x, y, acc):
    '''
    Convert the given screen coordinates to coordinates relative to
    the window that the given accessible is in.

    @param x: x screen coordinate to convert to window coordinate.
    @type  x: int
    @param y: y screen coordinate to convert to window coordinate.
    @type  y: int
    @param acc: accessible in the window relative to which the coordinates
                should be calculated.
    @type acc:  Atspi.Accessible
    @return: The (x, y) coordinates relative to the window that the accessible is in.
    @rtype: tuple(int, int)
    '''
    try:
      component_iface = acc.queryComponent()
    except NotImplementedError:
      return x, y

    acc_screen_extents = self.getScreenExtents(acc)
    acc_window_extents = component_iface.getExtents(pyatspi.WINDOW_COORDS)

    win_x = acc_window_extents.x - acc_screen_extents.x + x
    win_y = acc_window_extents.y - acc_screen_extents.y + y
    return (win_x, win_y)

  def getMousePosition(self):
    '''
    Get current mouse/pointer position in desktop/global coordinates.

    @return: The (x, y) coordinates of the mouse/pointer position.
    @rtype: tuple(int, int)
    '''
    display = Gdk.Display.get_default()
    seat = display.get_default_seat()
    pointer = seat.get_pointer()
    screen, x, y =  pointer.get_position()
    return x, y


class KWinWindowManager(WindowManager):
  '''
  WindowManager implementation that retrireves information from KWin
  via its scripting API.

  KWin API documentation: https://develop.kde.org/docs/plasma/kwin/api/
  '''

  def __init__(self):
    # assume that KWin has the same major version as KDE Plasma
    plasma_version_str = os.getenv('KDE_SESSION_VERSION')
    if plasma_version_str:
      self.kwin_version = int(plasma_version_str)
    else:
      # fall back to 5 for now, might be relevant when KWin is used in non-Plasma environment
      self.kwin_version = 5

  def _runKWinScript(self, kwin_script_path):
    '''
    Run KWin script at the given path that returns output in JSON format
    and return the script output converted to a Python datastructure.
    '''
    # load the script, the method returns the script number/ID
    session_bus = dbus.SessionBus()
    scripting_dbus_object = session_bus.get_object('org.kde.KWin', '/Scripting')
    scripting_interface = dbus.Interface(scripting_dbus_object, 'org.kde.kwin.Scripting')
    script_id = scripting_interface.loadScript(kwin_script_path)

    # remember time before running the script, for retrieving relevant journal content below
    start_time = datetime.datetime.now()

    # run the script
    # DBus object path for script has changed between KWin 5 and 6
    if self.kwin_version >= 6:
      script_object_path = '/Scripting/Script' + str(script_id)
    else:
      script_object_path = '/' + str(script_id)
    script_dbus_object = session_bus.get_object('org.kde.KWin', script_object_path)
    script_interface = dbus.Interface(script_dbus_object, 'org.kde.kwin.Script')
    script_interface.run()

    # currently, no DBus signals are created when the script prints the relevant information,
    # see https://bugs.kde.org/show_bug.cgi?id=477069
    # and https://bugs.kde.org/show_bug.cgi?id=392840
    # and https://bugs.kde.org/show_bug.cgi?id=445058
    # As a workaround, retrieve the script output from the journal instead,
    # s.a. discussion in https://bugs.kde.org/show_bug.cgi?id=445058
    comm = 'kwin_' + os.getenv('XDG_SESSION_TYPE')
    journalctl_output = subprocess.run('journalctl _COMM=' + comm + ' --output=cat --since "' + str(start_time) + '"',
                                       capture_output=True, shell=True).stdout.decode().rstrip().split("\n")
    lines = [line.lstrip("js: ") for line in journalctl_output]
    data_json = '\n'.join(lines)
    data = json.loads(data_json)

    # unload/unregister script again
    script_interface.stop()

    return data

  def getWindowInfos(self):
    # Retrieve information on all windows from KWin via the KWin scripting API
    #
    # See the JavaScript script used for details on the script output that gets
    # processed here
    if self.kwin_version >= 6:
      kwin_script_path = os.path.join(sys.prefix, 'share', 'accerciser', 'kwin-scripts', 'kwin6-retrieve-window-infos.js')
    else:
      kwin_script_path = os.path.join(sys.prefix, 'share', 'accerciser', 'kwin-scripts', 'kwin5-retrieve-window-infos.js')

    window_infos = []
    try:
      window_data = self._runKWinScript(kwin_script_path)
      for win in window_data:
        buffer_geometry = WindowGeometry(win["bufferGeometry.x"], win["bufferGeometry.y"],
                                         win["bufferGeometry.width"], win["bufferGeometry.height"])
        win_info = WindowInfo(win["caption"], win["geometry.x"], win["geometry.y"],
                              win["geometry.width"], win["geometry.height"],
                              stacking_index=win["stackingOrder"],
                              on_current_workspace=win["isOnCurrentWorkspace"],
                              buffer_geometry=buffer_geometry,
                              app_id=win["desktopFileName"], process_id=win["pid"])
        window_infos.append(win_info)
    except Exception:
      pass

    return window_infos

  def supportsScreenCoords(self, acc):
    # never query screen/desktop coordinates from AT-SPI, but always
    # use KWin's window position
    return False


  def getMousePosition(self):
    try:
      kwin_script_path = os.path.join(sys.prefix, 'share', 'accerciser', 'kwin-scripts', 'kwin-retrieve-mouse-pointer-pos.js')
      data = self._runKWinScript(kwin_script_path)
      return data["mouse-pointer-pos.x"], data["mouse-pointer-pos.y"]
    except:
      return 0, 0

  def getApplicationIcon(self, app):
    return self.getApplicationIconViaAppID(app)

class GnomeShellWindowManager(WindowManager):
  '''
  WindowManager implementation that retrieves information from
  Accerciser's GNOME Shell extension that queries the information
  from Mutter and provides them via a DBus service.
  '''

  def _getWindowData(self):
    '''
    Query window information from the Accerciser GNOME Shell extension
    via DBus.
    '''
    session_bus = dbus.SessionBus()
    accerciser_dbus_object = session_bus.get_object('org.gnome.accerciser.Accerciser', '/org/gnome/accerciser/Accerciser')
    accerciser_interface = dbus.Interface(accerciser_dbus_object, 'org.gnome.accerciser.Accerciser')
    data_json = accerciser_interface.GetWindowInfos()
    data = json.loads(data_json)
    return data

  def getWindowInfos(self):
    window_infos = []
    try:
      window_data = self._getWindowData()
      for win in window_data:
        buffer_geometry = WindowGeometry(win["bufferGeometry.x"], win["bufferGeometry.y"],
                                         win["bufferGeometry.width"], win["bufferGeometry.height"])

        # try both of "gtk_application_id" and "sandboxed_app_id" to retrieve an app ID
        app_id = win["gtk_application_id"]
        if not app_id:
          app_id = win["sandboxed_app_id"]

        win_info = WindowInfo(win["caption"], win["geometry.x"], win["geometry.y"],
                              win["geometry.width"], win["geometry.height"],
                              on_current_workspace=win["isOnCurrentWorkspace"],
                              buffer_geometry=buffer_geometry,
                              app_id=app_id, process_id=win["pid"])
        window_infos.append(win_info)
    except Exception:
      pass

    return window_infos

  def supportsScreenCoords(self, acc):
    # never query screen/desktop coordinates from AT-SPI, but always
    # use window position retrieved from Mutter/GNOME Shell extension
    return False

  def getMousePosition(self):
    try:
      session_bus = dbus.SessionBus()
      accerciser_dbus_object = session_bus.get_object('org.gnome.accerciser.Accerciser', '/org/gnome/accerciser/Accerciser')
      accerciser_interface = dbus.Interface(accerciser_dbus_object, 'org.gnome.accerciser.Accerciser')
      x, y = accerciser_interface.GetMousePosition()
      return x, y
    except Exception:
      return 0, 0

  def getApplicationIcon(self, app):
    return self.getApplicationIconViaAppID(app)

class WnckWindowManager(WindowManager):

  def _runWnckScript(self, args):
    '''
    Run Python script that uses Wnck to retrieve window information
    with the given arguments and return the script (JSON) output converted
    to a Python data structure.
    '''
    try:
      # run script with env var GDK_BACKEND=x11 because Wnck only works on X11/XWayland
      env_vars = os.environ
      env_vars['GDK_BACKEND'] = 'x11'
      wnck_script_path = os.path.join(sys.prefix, 'share', 'accerciser', 'wnck-window-infos.py')
      all_args = [wnck_script_path] + args
      script_output = subprocess.run(all_args,
                                     capture_output=True,
                                     check=True,
                                     env=env_vars
                                    ).stdout.decode()
      data = json.loads(script_output)
      return data
    except Exception:
      return None

  def getWindowInfos(self):
    window_data = self._runWnckScript(['window-infos'])
    if not window_data:
      return []

    window_infos = []
    for win in window_data:
      win_info = WindowInfo(win["caption"], win["geometry.x"], win["geometry.y"],
                            win["geometry.width"], win["geometry.height"],
                            stacking_index=win["stackingOrder"],
                            on_current_workspace=win["isOnCurrentWorkspace"],
                            process_id=win["pid"])
      window_infos.append(win_info)
    return window_infos

  def getApplicationIcon(self, app):
    window_names = [app.get_child_at_index(i).get_name() for i in range(0, app.get_child_count())]
    response = self._runWnckScript(['icon'] + window_names)
    if not response or not ('pixels' in response):
      return None

    try:
      pixbuf = GdkPixbuf.Pixbuf.new_from_data(
        bytes.fromhex(response['pixels']),
        GdkPixbuf.Colorspace.RGB,
        response['has_alpha'],
        response['bits_per_sample'],
        response['width'],
        response['height'],
        response['rowstride']
      )
      return pixbuf
    except:
      return None



'''
Defines a mapping from widget roles to icons representing them.

@author: Peter Parente
@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

import gi

from gi.repository import Gtk as gtk
from gi.repository import GdkPixbuf
from gi.repository import GObject

import sys, os, glob

from . import window_manager

from pyatspi.constants import *

ICONS_PATH = os.path.join(sys.prefix, 'share',
                          'accerciser', 'pixmaps', 'hicolor', '22x22')

if not os.path.exists(ICONS_PATH):
   ICONS_PATH = os.path.join(os.path.split(os.getcwd())[0], 'pixmaps')

def getIcon(acc):
  '''
  Loads an icon for the given application or accessible widget. Tries to use
  the current theme or the window manager to get application icons. Uses own
  icons for widgets.
  '''
  theme = gtk.IconTheme.get_default()
  try:
    role_name = acc.getRoleName()
    role = acc.getRole()
    if role_name == 'application':
      # try the theme first
      try:
        return theme.load_icon(acc.name, 24, gtk.IconLookupFlags.USE_BUILTIN)
      except GObject.GError:
        pass

      # then try the WindowManager
      win_manager = window_manager.get_window_manager()
      icon = win_manager.getApplicationIcon(acc)
      return icon
    else:
      name = role_name.replace(' ', '')
      try:
        fn = os.path.join(ICONS_PATH, '%s.png' % name)
        return GdkPixbuf.Pixbuf.new_from_file(fn)
      except GObject.GError:
        pass
  except Exception as e:
    pass
  fn = os.path.join(ICONS_PATH, 'filler.png')
  return GdkPixbuf.Pixbuf.new_from_file(fn)



'''
Defines the main program classes for creating the GUI, drawing the blinking
selection rectangle, and maintaining information about the currently selected
accessible.

@author: Peter Parente
@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

from gi.repository import Gio as gio
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import GLib
from gi.repository import Atk as atk

import os, sys, locale
from .icons import getIcon
import os
from .bookmarks import BookmarkStore
from .accessible_treeview import *
from .node import Node
from .plugin import PluginManager
from .plugin import PluginView
from .tools import Tools
from .i18n import _, N_
from .hotkey_manager import HotkeyManager, HotkeyTreeView
from .about_dialog import AccerciserAboutDialog
from .prefs_dialog import AccerciserPreferencesDialog
from .main_window import AccerciserMainWindow
from . import menus, window_manager

class Main(gtk.Application, Tools):
  '''
  Class for the main accerciser window.
  '''
  COL_ACC = 4
  COL_FILLED = 2

  def __init__(self, *args, **kwargs):
    gtk.Application.__init__(self, *args, **kwargs,
                             application_id="org.gtk.accerciser")
    self.window = None

  def do_startup(self):
    '''
    gtk.Application callback that gets called when application starts.
    Set up the application.
    '''
    gtk.Application.do_startup(self)

    # mark the root of this window with its PID so we can easily identify it
    # as this app
    root_atk = atk.get_root()
    root_atk.set_description(str(os.getpid()))

    self.node = Node(window_manager.get_window_manager())
    self.set_menubar(menus.main_menu)

  def do_activate(self):
    '''
    gtk.Application callback when application gets launched
    by desktop environment.
    Set up and show the ApplicationWindow.
    Gets references to important widgets, establishes event handlers,
    configures the tree view, and initializes the tree with the contents of the
    desktop.
    '''
    if self.window:
        self.window.present()
        return

    self.window = AccerciserMainWindow(application=self, node=self.node)
    self.window.connect('delete-event', self._onDeleteEvent)
    self.window.connect('destroy', self._onQuit)

    # Start hotkey manager
    self.hotkey_manager = HotkeyManager()
    self.bookmarks_store = BookmarkStore(self.node, self, self.window)

    # load plugins
    self.plugin_manager = \
        PluginManager(self, self.node, self.hotkey_manager,
                      self.window.pluginview1, self.window.pluginview2)

    # connect signal handlers and show the GUI in its initial state
    self.window.show_all()

    menu_items = [
        (menus.file_menu, 'quit', 'application-exit', _('_Quit'),
         '<control>q', self._onQuit),
        (menus.edit_menu, 'preferences', 'preferences-system', _('_Preferences…'),
         '<control>p', self._onShowPreferences),
        (menus.help_menu, 'contents', 'help-browser', _('_Contents'),
         'F1', self._onHelp),
        (menus.help_menu, 'about', 'help-about', _('_About'),
         None, self._onAbout)]

    self.addMenuItems(menu_items)

    self.last_focused = None
    self.window.show_all()

    try:
      # async is a reserved keyword in Python 3.7+, so we pass the args as dict
      pyatspi.Registry.start(**{'async': True, 'gil': False})
    except KeyboardInterrupt:
      self._shutDown()

  def addMenuItem(self, menu, name, icon_name, label, accel, callback):
    '''
    Create menu item in the given menu and return the menu item and its action.

    @param menu: The menu into which the item should be inserted.
    @type menu: gio.Menu
    @param name: Name for the action
    @type name: string
    @param icon_name: Name of the icon to use.
    @type icon_name: string
    @param label: Label for the menu entry.
    @type label: string
    @param accel: Keyboard accelerator for the action, or None.
    @type accel: string
    @param callback: Callback that gets called when the action gets triggered.
    @type callback: method

    @return: Menu item and action.
    @rtype: tuple(gio.MenuItem, gio.SimpleAction)
    '''
    action_name = 'app.' + name
    menu_item = gio.MenuItem.new(label, action_name)
    icon = gio.ThemedIcon.new(icon_name)
    menu_item.set_icon(icon)
    menu.append_item(menu_item)
    action = gio.SimpleAction.new(name, None)
    action.connect('activate', callback)
    if accel:
      self.set_accels_for_action(action_name, [accel])
    self.add_action(action)
    return menu_item, action

  def addMenuItems(self, menu_items):
    '''
    Add menu items. The parameter is a list of tuples containing
    all parameters that L{addMenuItem} takes.
    See L{addMenuItem} for more details.

    @param menu_items
    @type list(tuple(gio.Menu, string, string, string, string, method))
    '''
    for menu, name, icon_name, label, accel, callback in menu_items:
      self.addMenuItem(menu, name, icon_name, label, accel, callback)

  def _shutDown(self):
    '''
    Cleans up any object instances that need explicit shutdown.
    '''
    self.plugin_manager.close()

  def _onQuit(self, obj, data=None):
    '''
    Quits the app.

    @param obj: The object that emitted the signal that this callback caught.
    @type obj: L{gtk.Widget}
    '''
    self._shutDown()
    pyatspi.Registry.stop()
    self.quit()

  def _onAbout(self, action, data=None):
    '''
    Shows the about dialog.

    @param widget: The widget that emitted the signal that this callback caught.
    @type widget: L{gtk.Widget}
    '''
    about = AccerciserAboutDialog()
    about.set_transient_for(self.window)
    about.run()

  def _onHelp(self, action, page=""):
    '''
    Shows the help dialog.

    @param widget: The widget that emitted the signal that this callback caught.
    @type widget: L{gtk.Widget}
    '''
    uri = "help:accerciser"
    if page:
        uri += "/%s" % page
    gtk.show_uri_on_window(self.window,
                           uri,
                           gtk.get_current_event_time())
    return True

  def _onShowPreferences(self, action, data=None):
    '''
    Shows the preferences dialog.

    @param widget: The widget that emitted the signal that this callback caught.
    @type widget: L{gtk.Widget}
    '''
    plugins_view = self.plugin_manager.View()
    hotkeys_view = HotkeyTreeView(self.hotkey_manager)
    dialog = AccerciserPreferencesDialog(plugins_view, hotkeys_view)
    dialog.set_transient_for(self.window)
    dialog.show_all()

  def _onDeleteEvent(self, obj, data=None):
    '''
    Handles when a delete-event is triggered from the main window.

    @param obj: The object that emitted the signal that this callback caught.
    @type obj: L{gtk.Widget}
    '''
    self.window.saveState()



'''
Defines convenience classes representing tree nodes and bag objects.

@author: Peter Parente
@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
import gi

from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository.Gio import Settings as GSettings
gi.require_version('Rsvg', '2.0')
from gi.repository import Rsvg as rsvg
import cairo
import pyatspi
import string
from .tools import ToolsAccessor, parseColorString

gsettings = GSettings.new('org.a11y.Accerciser')
BORDER_COLOR, BORDER_ALPHA = parseColorString(
  gsettings.get_string('highlight-border'))

FILL_COLOR, FILL_ALPHA  = parseColorString(
  gsettings.get_string('highlight-fill'))

HL_DURATION = int(gsettings.get_double('highlight-duration')*1000)

class Bag(object):
  '''
  Bag class for converting a dicionary to an object with attributes.
  '''
  def __init__(self, **kwargs):
    self.__dict__.update(kwargs)

  def __str__(self):
    return ', '.join(list(vars(self).keys()))

class Node(GObject.GObject, ToolsAccessor):
  '''
  Node class that contains convient references to accessibility information
  for the currently selected node. A L{Node} instance will emit an
  'accessible-changed' signal when L{update} is called with a new accessible.

  @ivar desktop: The desktop accessible. It holds references to all the
  application L{Accessibility.Accessible}s
  @type desktop: L{Accessibility.Accessible}
  @ivar acc: The currently selected accessible.
  @type acc: L{Accessibility.Accessible}
  @ivar window_manager: The window manager.
  @type window_manager: L{WindowManager}
  '''
  __gsignals__ = {'accessible-changed' :
                  (GObject.SignalFlags.RUN_FIRST,
                   None,
                   (GObject.TYPE_PYOBJECT,))}
  def __init__(self, window_manager):
    self.desktop = pyatspi.Registry.getDesktop(0)
    self.acc = None
    self.tree_path = None
    self.window_manager = window_manager
    GObject.GObject.__init__(self)

  def update(self, acc):
    '''
    Updates the information in this node for the given accessible including
    a reference to the accessible. Also emit the
    'accessible-changed' signal.

    @param acc: An accessible.
    @type acc: L{Accessibility.Accessible}
    '''
    if not acc or self.isMyApp(acc):
      return
    self.acc = acc
    self.tree_path = None
    if acc != self.desktop:
        # Don't highlight the entire desktop, it gets annoying.
        self.highlight()
    self.emit('accessible-changed', acc)

  def updateToPath(self, app_name, path):
    '''
    Update the node with a new accessible by providing a tree path
    in an application.

    @param app_name: Application name.
    @type app_name: string
    @param path: The accessible path in the application.
    @type path: list of integer
    '''
    acc = pyatspi.findDescendant(
      self.desktop,
      lambda x: x.name.lower()==app_name.lower(),
      breadth_first=True)
    if acc is None: return
    while path:
      child_index = path.pop(0)
      try:
        acc = acc[child_index]
      except IndexError:
        return
    self.update(acc)

  def highlight(self):
    extents = self.window_manager.getScreenExtents(self.acc)
    if extents is None or \
          0 in (extents.width, extents.height) or \
          -0x80000000 in (extents.x, extents.y):
      return
    ah = _HighLight(extents.x, extents.y,
                    extents.width, extents.height,
                    FILL_COLOR, FILL_ALPHA, BORDER_COLOR, BORDER_ALPHA,
                    2.0, 0)
    ah.highlight(HL_DURATION)

class _HighLight(gtk.Window):
  '''
  Highlight box class. Uses compositing when available. When not, it does
  transparency client-side.
  '''
  _svg = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg xmlns="http://www.w3.org/2000/svg">
  <rect
       style="fill:$fill;fill-opacity:$fill_opacity;fill-rule:evenodd;stroke:$stroke;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:$stroke_opacity"
       id="highlight"
       width="$width"
       height="$height"
       x="$x"
       y="$y"
       rx="2"
       ry="2" />
</svg>
"""
  def __init__(self, x, y, w, h,
               fill_color, fill_alpha,
               stroke_color, stroke_alpha,
               stroke_width, padding=0):

    # Initialize window.
    gtk.Window.__init__(self, type=gtk.WindowType.POPUP)

    # Normalize position for stroke and padding.
    self.x, self.y = x - padding, y - padding
    self.w, self.h = w + padding*2, h + padding*2

    # Determine if we are compositing.
    screen = self.get_screen()
    self._composited = screen.is_composited()
    if self._composited:
      # Prepare window for transparency.
      visual = screen.get_rgba_visual()
      self.set_visual(visual)

    # Place window, and resize it, and set proper properties.
    self.set_app_paintable(True)
    self.set_decorated(False)
    self.set_keep_above(True)
    self.move(self.x, self.y)
    self.resize(self.w, self.h)
    self.set_accept_focus(False)
    self.set_sensitive(False)

    # Create SVG with given parameters.
    offset = stroke_width/2.0
    self.svg = string.Template(self._svg).substitute(
      x=offset, y=offset,
      width=int(self.w - stroke_width), height=int(self.h - stroke_width),
      fill=fill_color,
      stroke_width=stroke_width,
      stroke=stroke_color,
      fill_opacity=fill_alpha,
      stroke_opacity=stroke_alpha)

    # Connect "draw"
    self.connect("draw", self._onDraw)

  def highlight(self, duration=500):
    if duration > 0:
      GLib.timeout_add(duration, lambda w: w.destroy(), self)
      self.show_all()
    else:
      self.destroy()

  def _onDraw(self, widget, cr):
    svgh = rsvg.Handle()
    try:
      svgh.write(bytes(self.svg, "utf-8"))
    except (GObject.GError, KeyError, ValueError) as ex:
      print ('Error reading SVG for display: %s\r\n%s', ex, self.svg)
      svgh.close()
      return
    svgh.close()

    if not self._composited:
      cairo_operator = cairo.OPERATOR_OVER
    else:
      cairo_operator = cairo.OPERATOR_SOURCE
    cr.set_source_rgba(1.0, 1.0, 1.0, 0.0)
    cr.set_operator(cairo_operator)

    rect = rsvg.Rectangle()
    rect.x = 0
    rect.y = 0
    rect.width = self.w
    rect.height = self.h

    svgh.render_document(cr, rect)
    del svgh

if __name__ == "__main__":
    hl = _HighLight(200, 200, 200, 200, '#ff0000',
                    0.5, '#ff0000', 0.996108949416, 8.0, 0)
    hl.highlight(2000)
    gtk.main()




'''
Defines a base class having tools common to the core and the plugins.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
import os
import pickle
import weakref

import traceback
import functools

class Tools(object):
  '''
  A class with some common methods that more than a few classes will need.

  @ivar pid: Process ID of current Accerciser instance.
  @type pid: integer
  '''

  def isMyApp(self, acc):
    '''
    Checks if a given L{Accessibility.Accessible} belongs to this current
    app instance. This is useful for avoiding recursion and such.

    @param acc: The given L{Accessibility.Accessible} to check.
    @type acc: L{Accessibility.Accessible}
    @return: True if L{acc} is a member of current app instance.
    @rtype: boolean
    '''
    if not hasattr(self, 'pid'):
      self.pid = os.getpid()

    if not acc:
      return False
    try:
      app = acc.get_application()
      # check whether PID is set as description, set in Main.do_startup
      return app.get_description() == str(self.pid)
    except Exception as e:
      return False


class Proxy(object):
  '''
  Our own proxy object which enables weak references to bound and unbound
  methods and arbitrary callables. Pulls information about the function,
  class, and instance out of a bound method. Stores a weak reference to the
  instance to support garbage collection.
  '''
  def __init__(self, cb):
    try:
      try:
        self.inst = weakref.ref(cb.__self__)
      except TypeError:
        self.inst = None
      self.func = cb.__func__
      self.klass = cb.__self__.__class__
    except AttributeError:
      self.inst = None
      self.func = cb.__func__
      self.klass = None

  def __call__(self, *args, **kwargs):
    '''
    Proxy for a call to the weak referenced object. Take arbitrary params to
    pass to the callable.

    @raise ReferenceError: When the weak reference refers to a dead object
    '''
    if self.inst is not None and self.inst() is None:
      return
    elif self.inst is not None:
      # build a new instance method with a strong reference to the instance
      mtd = self.func.__get__(self.inst(), self.klass)
    else:
      # not a bound method, just return the func
      mtd = self.func
    # invoke the callable and return the result
    return mtd(*args, **kwargs)

  def __eq__(self, other):
    '''
    Compare the held function and instance with that held by another proxy.

    @param other: Another proxy object
    @type other: L{Proxy}
    @return: Whether this func/inst pair is equal to the one in the other proxy
      object or not
    @rtype: boolean
    '''
    try:
      return self.func == other.func and self.inst() == other.inst()
    except Exception:
      return False

  def __ne__(self, other):
    '''
    Inverse of __eq__.
    '''
    return not self.__eq__(other)

def parseColorString(color_string):
  '''
  Parse a string representation of a 24-bit color, and a 8 bit alpha mask.

  @param color_string: String in the format: #rrbbggaa.
  @type color_string: string

  @return: A color string in the format of #rrggbb, and an opacity value
  of 0.0 to 1.0
  @rtype: tuple of string and float.
  '''
  return color_string[:-2], int(color_string[-2:], 16)/255.0

def getTreePathBoundingBox(treeview, path, col):
  '''
  Get bounding box of given tree path.
  '''
  gdkwindow = treeview.window
  x, y = treeview.allocation.x, treeview.allocation.y
  while gdkwindow:
    window_x, window_y = gdkwindow.get_position()
    x += window_x
    y += window_y
    gdkwindow = gdkwindow.get_parent()
  rect = treeview.get_cell_area(path, col)
  rect.x, rect.y = treeview.tree_to_widget_coords(rect.x, rect.y)
  rect.x += x
  rect.y += y
  return rect

def logException(func):
  '''
  Handle (and log) the exceptions that are coming from plugins
  '''
  @functools.wraps(func)
  def newfunc(*args, **kwargs):
    # use Exception otherwise KeyboardInterrupt won't get through
    try:
      return func(*args, **kwargs)
    except Exception:
      traceback.print_exc()
  return newfunc

class ToolsAccessor(Tools):
  '''
  By following the recommendation on
  https://bugzilla.gnome.org/show_bug.cgi?id=723081#c4, this Accessor allows us
  to wrap every plugin's method and in order to catch all possible exceptions
  and print them appropiately.
  '''
  def __init__(self, plugin):
    self.plugin = plugin

    @logException
    def method(self):
      return self.plugin.method()



'''
AT-SPI interface viewer plugin.

@author: Eitan Isaacson
@organization: Mozilla Foundation
@copyright: Copyright (c) 2007 Mozilla Foundation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

import gi

from gi.repository import Atspi
from gi.repository import Gtk as gtk
from gi.repository import GdkPixbuf
from gi.repository import Pango
from gi.repository.GLib import markup_escape_text

import pyatspi
import os.path

from accerciser.plugin import ViewportPlugin
from accerciser.icons import getIcon
from accerciser import node
from accerciser.i18n import _, N_, DOMAIN
from xml.dom import minidom

UI_FILE = os.path.join(os.path.dirname(__file__),
                       'interface_view.ui')

class InterfaceViewer(ViewportPlugin):
  '''
  Interface Viewer plugin class.

  @ivar label_role: Top plugin label that displays the role name and
  the accessible name.
  @type label_role: gtk.Label
  @ivar sections: List of L{_InterfaceSection} instances.
  @type sections: list
  '''
  # Translators: this is a plugin name
  plugin_name = N_('Interface Viewer')
  plugin_name_localized = _(plugin_name)
  # Translators: this is a plugin description
  plugin_description = N_('Allows viewing of various interface properties')

  def init(self):
    '''
    Intialize plugin.
    '''
    # HACK: Put all the callbacks in this class.
    dom = minidom.parse(UI_FILE)
    callbacks= set([signal.getAttribute('handler') \
                      for signal in dom.getElementsByTagName('signal')])
    del dom

    ui_xml = gtk.Builder()
    ui_xml.set_translation_domain(DOMAIN)
    ui_xml.add_from_file(UI_FILE)
    frame = ui_xml.get_object('iface_view_frame')
    self.label_role = ui_xml.get_object('label_role')
    self.plugin_area.add(frame)
    self.sections = [
      _SectionAccessible(ui_xml, self.node),
      _SectionAction(ui_xml, self.node),
      _SectionApplication(ui_xml, self.node),
      _SectionComponent(ui_xml, self.node),
      _SectionDocument(ui_xml, self.node),
      _SectionHyperlink(ui_xml, self.node),
      _SectionHypertext(ui_xml, self.node),
      _SectionImage(ui_xml, self.node),
      _SectionSelection(ui_xml, self.node),
      _SectionStreamableContent(ui_xml, self.node),
      _SectionTable(ui_xml, self.node),
      _SectionTableCell(ui_xml, self.node),
      _SectionText(ui_xml, self.node),
      _SectionValue(ui_xml, self.node),
      _SectionCollection(ui_xml, self.node),
      _SectionDesktop(ui_xml, self.node),
      _SectionLoginHelper(ui_xml, self.node)]

    # HACK: Add callbacks to this class.
    for cb in callbacks:
      for section in self.sections:
        method = getattr(section, cb, None)
        if not method: continue
        setattr(self, cb, method)

    ui_xml.connect_signals(self)

    # Mark all expanders with no associated section classes as unimplemented
    implemented_expanders = [s.expander for s in self.sections]
    vbox_ifaces = ui_xml.get_object('vbox_ifaces')

    for expander in vbox_ifaces.get_children():
      if expander not in implemented_expanders:
        iface_name = \
          expander.get_label().lower().replace(' ', '').replace('_', '')
        section = _InterfaceSection(ui_xml, self.node, iface_name)
        section.disable()

    pyatspi.Registry.registerEventListener(
        self.onAccNameChanged, 'object:property-change:accessible-name')

  def onAccNameChanged(self, event):
    '''
    Listener for accessible name changes, if it is ours, change the name.

    @param event: 'object:property-change:accessible-name' event.
    @type acc: Accessibility.Event
    '''
    if event.source != self.node.acc:
      return

    role = self.node.acc.getRoleName()
    name = self.node.acc.name
    if name:
      role_name = '%s: %s' % (role, name)
    else:
      role_name = role
    self.label_role.set_markup('<b>%s</b>' % markup_escape_text(role_name))

  def onAccChanged(self, acc):
    '''
    Method that is invoked when the main accessible selection s changed.

    @param acc: New accessible
    @type acc: Accessibility.Accessible
    '''
    role = acc.getRoleName()
    name = acc.name
    if name:
      role_name = '%s: %s' % (role, name)
    else:
      role_name = role
    self.label_role.set_markup('<b>%s</b>' % markup_escape_text(role_name))
    interfaces = pyatspi.listInterfaces(acc)
    for section_obj in self.sections:
      section_obj.disable()
      if section_obj.interface_name in interfaces:
        section_obj.enable(acc)

class _InterfaceSection(object):
  '''
  An abstract class that defines the interface for interface sections.

  @cvar interface_name: Name of interface this section is for.
  @type interface_name: string

  @ivar node: Application-wide L{Node}.
  @type node: L{Node}
  @ivar expander: The section expander widget.
  @type expander: gtk.Expander
  @ivar event_listeners: List of client and event pairs that are
  registered by this section. They are typically registered an population time,
  and alwas deregistered on L{clearUI}.
  @type event_listeners: list
  '''
  interface_name = None
  def __init__(self, ui_xml, node, interface_name=None):
    '''
    Initialize section object. and call init() for derived classes.

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    @param node: Application-wide node of selected accessible.
    @type node: L{Node}
    @param interface_name: Override default interface name.
    @type interface_name: string
    '''
    self.interface_name = interface_name or self.interface_name
    self.node = node
    self.expander = \
        ui_xml.get_object('expander_%s' % self.interface_name.lower())
    self._setExpanderChildrenSensitive(False)
    self.event_listeners = []
    self.init(ui_xml)

  def init(self, ui_xml):
    '''
    Abtract method for initializing section-specific code.

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    pass

  def enable(self, acc):
    '''
    Make section sensitive and populate it.

    @param acc: Accessible to use for information when populating section.
    @type acc: Accessibility.Accessible
    '''
    self._setExpanderChildrenSensitive(True)
    self.populateUI(acc)

  def populateUI(self, acc):
    '''
    Abstract method for section specific populating code.

    @param acc: Accessible to use for information when populating section.
    @type acc: Accessibility.Accessible
    '''
    pass

  def disable(self):
    '''
    Disable section, make insensitive.
    '''
    self._setExpanderChildrenSensitive(False)
    for client, event_names in self.event_listeners:
      pyatspi.Registry.deregisterEventListener(client, *event_names)
    self.clearUI()

  def clearUI(self):
    '''
    Abstract method for section-specific cleanup.
    '''
    pass

  def _setExpanderChildrenSensitive(self, sensitive, expander=None):
    '''
    Convenience method for making the expander's children insensitive.
    We don't want tomake the expander itself insensitive because the user might
    still want to keep it open or close it when it is disabled.

    @param sensitive: True for sensitive.
    @type sensitive: boolean
    @param expander: Expander widget. Uses instances expander by default.
    @type expander: gtk.Expander
    '''
    expander = expander or self.expander
    label = expander.get_label_widget()
    label_text = label.get_label()
    not_implemented_str = _('(not implemented)')

    if sensitive:
      label_text = label_text.replace(not_implemented_str, '')
      label_text = label_text.strip(' ')
    elif not not_implemented_str in label_text:
      label_text = label_text + ' ' + not_implemented_str
    label.set_label(label_text)
    for child in expander.get_children():
      child.set_sensitive(sensitive)

  def _isSelectedInView(self, selection):
    '''
    Convenience method for determining if a given treeview selection has any
    selected nodes.

    @param selection: Selection to check.
    @type selection: gtk.TreeSelection

    @return: True is something is selected
    @rtype: boolean
    '''
    model, rows = selection.get_selected_rows()
    something_is_selected = bool(rows)
    return something_is_selected

  def _onViewSelectionChanged(self, selection, *widgets):
    '''
    Convenience callback for selection changes. Useful for setting given
    widgets to be sensitive only when something is selected, for example
    action buttons.

    @param selection: The selection object that triggered the callback.
    @type selection: gtk.TreeSelection
    @param widgets: list of widgets that should be made sensitive/insensitive
    on selection changes.
    @type widgets: list of gtk.Widget
    '''
    for widget in widgets:
      widget.set_sensitive(self._isSelectedInView(selection))

  def registerEventListener(self, client, *event_names):
    pyatspi.Registry.registerEventListener(client, *event_names)
    self.event_listeners.append((client, event_names))

class _SectionAccessible(_InterfaceSection):
  '''
  A class that populates an Accessible interface section.

  @ivar states_model: Model for accessible states.
  @type states_model: gtk.ListStore
  @ivar relations_view: Tree view for accessible relations.
  @type relations_view: gtk.TreeView
  @ivar relations_model: Model for accessible relations.
  @type relations_view: gtk.TreeStore
  @ivar attr_model: Model for accessible attributes.
  @type attr_model: gtk.ListStore
  @ivar show_button: Button to jump to the selected relation target object.
  @type show_button: gtk.Button
  '''

  interface_name = 'Accessible'

  def init(self, ui_xml):
    '''
    Initialization that is specific to the Accessible interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # Child count and description labels
    self.child_count_label = ui_xml.get_object('label_acc_child count')
    self.desc_label = ui_xml.get_object('label_acc_desc')
    self.help_text_label = ui_xml.get_object('label_acc_help_text')
    self.id_label = ui_xml.get_object('label_acc_id')
    self.process_id_label = ui_xml.get_object('label_process_id')

    # configure states tree view
    self.states_model = ui_xml.get_object('states_liststore')

    # configure relations tree view
    self.relations_view = ui_xml.get_object('relations_view')
    self.relations_model = ui_xml.get_object('relations_treestore')
    # preset the different bg colors
    style = self.relations_view.get_style_context()

    selection = self.relations_view.get_selection()
    self.show_button = ui_xml.get_object('button_relation_show')
    self.show_button.set_sensitive(self._isRelationTargetSelected())
    selection.connect('changed', self._onRelationsSelectionChanged)

    # configure accessible attributes tree view
    self.attr_model = ui_xml.get_object('accattrib_liststore')

  def populateUI(self, acc):
    '''
    Populate the Accessible section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''

    self.child_count_label.set_text(str(acc.childCount))
    self.desc_label.set_label(acc.description or _('(no description)'))
    try:
        self.help_text_label.set_label(acc.get_help_text() or _('(no help text)'))
    except:
        self.help_text_label.set_label(_('(no help text)'))
    try:
        self.id_label.set_label(acc.accessibleId)
    except:
        self.id_label.set_label(_('(no ID)'))
    self.process_id_label.set_text(str(acc.get_process_id()))

    states = [pyatspi.stateToString(s) for s in acc.getState().getStates()]
    states.sort()
    list(map(self.states_model.append, [[state] for state in states]))

    try:
      attribs = acc.getAttributes()
    except:
      pass
    else:
      for attr in attribs:
        name, value = attr.split(':', 1)
        self.attr_model.append([name, value])

    relations = acc.getRelationSet()
    for relation in relations:
      r_type_name = repr(relation.getRelationType()).replace('RELATION_', '')
      r_type_name = r_type_name.replace('_', ' ').lower().capitalize()
      iter = self.relations_model.append(
          None, [None,
                 markup_escape_text(r_type_name), -1,
                 False])
      for i in range(relation.getNTargets()):
        acc = relation.getTarget(i)
        self.relations_model.append(
            iter, [getIcon(acc),
                   markup_escape_text(acc.name), i,
                   True])
    self.relations_view.expand_all()

    self.registerEventListener(self._accEventDescriptionChanged,
                               'object:property-change:accessible-description')
    self.registerEventListener(self._accEventHelpTextChanged,
                               'object:property-change:accessible-help-text')
    self.registerEventListener(self._accEventState, 'object:state-changed')

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.relations_model.clear()
    self.states_model.clear()
    self.attr_model.clear()

  def _isRelationTargetSelected(self):
    '''
    Returns whether a relation target object is currently selected
    in the relations tree view.
    '''
    selection = self.relations_view.get_selection()
    model, iter = selection.get_selected()
    # target objects are children of the node displaying the relation type
    return model and iter and model.get_path(iter).get_depth() > 1

  def _onRelationsSelectionChanged(self, selection):
    '''
    Callback for selection change in the relations tree view.

    @param selection: The selection object that triggered the callback.
    @type selection: gtk.TreeSelection
    '''
    self.show_button.set_sensitive(self._isRelationTargetSelected())

  def _onRelationShow(self, relations_view, *more_args):
    '''
    Callback for row activation or button press. Selects the related
    accessible in the main application.

    @param relations_view: The relations treeview.
    @type relations_view: gtk.TreeView
    @param *more_args: More arguments that are provided by variuos types of
    signals, but we discard them all.
    @type *more_args: list
    '''
    selection = relations_view.get_selection()
    model, iter = selection.get_selected()
    if iter and model[iter][2] >= 0:
      path = model.get_path(iter)
      relations = self.node.acc.getRelationSet()
      acc = relations[path[0]].getTarget(model[iter][2])
      if acc:
        self.node.update(acc)

  def _accEventDescriptionChanged(self, event):
    '''
    Callback for accessible description changes.

    @param event: Event that triggered this callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return

    self.desc_label.set_label(self.node.acc.description or _('(no description)'))

  def _accEventHelpTextChanged(self, event):
    '''
    Callback for accessible help text changes.

    @param event: Event that triggered this callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return

    try:
        self.help_text_label.set_label(self.node.acc.get_help_text() or _('(no help text)'))
    except:
        self.help_text_label.set_label(_('(no help text)'))

  def _accEventState(self, event):
    '''
    Callback for accessible state changes. Repopulates the states model.

    @param event: Event that triggered this callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc == event.source:
      self.states_model.clear()
      try:
        states = [pyatspi.stateToString(s) for s in \
                    self.node.acc.getState().getStates()]
      except LookupError:
        # Maybe we got a defunct state, in which case the object is diseased.
        states = []
      states.sort()
      list(map(self.states_model.append, [[state] for state in states]))


class _SectionAction(_InterfaceSection):
  '''
  A class that populates an Action interface section.

  @ivar actions_model: Model for accessible states.
  @type actions_model: gtk.ListStore
  @ivar action_selection: Current selection of actions tree view.
  @type action_selection: gtk.TreeSelection
  '''
  interface_name = 'Action'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Action interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure actions tree view
    treeview = ui_xml.get_object('treeview_action')
    self.actions_model = treeview.get_model()
    self.action_selection = treeview.get_selection()
    show_button = ui_xml.get_object('button_action_do')
    show_button.set_sensitive(self._isSelectedInView(self.action_selection))
    self.action_selection.connect('changed',
                                  self._onViewSelectionChanged, show_button)

  def populateUI(self, acc):
    '''
    Populate the Action section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ai = acc.queryAction()
    for i in range(ai.nActions):
      self.actions_model.append([i, ai.getName(i),
                                 ai.getDescription(i),
                                 ai.getKeyBinding(i)])

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.actions_model.clear()

  def _onActionRowActivated(self, treeview, path, view_column):
    '''
    Callback for row activation in action treeview. Performs actions.

    @param treeview: Actions tree view.
    @type treeview: gtk.TreeView
    @param path: Path of activated role.
    @type path: tuple
    @param view_column: The column that was clicked
    @type view_column: integer
    '''
    action_num = self.actions_model[path][0]
    ai = self.node.acc.queryAction()
    ai.doAction(action_num)

  def _onActionClicked(self, button):
    '''
    Callback for "do action" button. Performs action of currently selected row.

    @param button: The button that was pressed.
    @type button: gtk.Button
    '''
    actions_model, iter = self.action_selection.get_selected()
    action_num = actions_model[iter][0]
    ai = self.node.acc.queryAction()
    ai.doAction(action_num)

class _SectionApplication(_InterfaceSection):
  '''
  A class that populates an Application interface section.

  @ivar label_id: Label that displays application id info.
  @type label_id: gtk.Label
  @ivar label_tk: Label for toolkit name.
  @type label_tk: gtk.Label
  @ivar label_version: Label for toolkit version.
  @type label_version: gtk.Label
  '''
  interface_name = 'Application'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Application interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.label_id = ui_xml.get_object('label_app_id')
    self.label_tk = ui_xml.get_object('label_app_tk')
    self.label_version = ui_xml.get_object('label_app_version')

  def populateUI(self, acc):
    '''
    Populate the Application section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ai = acc.queryApplication()
    self.label_id.set_text(repr(ai.id))
    self.label_tk.set_text(ai.toolkitName)
    self.label_version.set_text(ai.version)

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.label_id.set_text('')
    self.label_tk.set_text('')
    self.label_version.set_text('')

class _SectionComponent(_InterfaceSection):
  '''
  A class that populates a Component interface section.

  @ivar label_posrelwin: Window relative position label
  @type label_posrelwin: gtk.Label
  @ivar label_posrelparent: Parent relative position label
  @type posrelparent: gtk.Label
  @ivar label_posabs: Absolute position label
  @type label_posabs: gtk.Label
  @ivar label_layer: Layer label
  @type label_layer: gtk.Label
  @ivar label_zorder: Z-order label
  @type label_zorder: gtk.Label
  @ivar label_alpha: Alpha label
  @type label_alpha: gtk.Label
  '''
  interface_name = 'Component'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Component interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.label_posabs = ui_xml.get_object('absolute_position_label')
    self.label_posrelwin = ui_xml.get_object('window_relative_position_label')
    self.label_posrelparent = ui_xml.get_object('parent_relative_position_label')
    self.label_size = ui_xml.get_object('size_label')
    self.label_layer = ui_xml.get_object('layer_label')
    self.label_zorder = ui_xml.get_object('zorder_label')
    self.label_alpha = ui_xml.get_object('alpha_label')

  def populateUI(self, acc):
    '''
    Populate the Component section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ci = acc.get_component()
    bbox = ci.get_extents(Atspi.CoordType.SCREEN)
    self.label_posabs.set_text('%d, %d' % (bbox.x, bbox.y))
    self.label_size.set_text('%dx%d' % (bbox.width, bbox.height))
    bbox = ci.get_extents(Atspi.CoordType.WINDOW)
    self.label_posrelwin.set_text('%d, %d' % (bbox.x, bbox.y))
    bbox = ci.get_extents(Atspi.CoordType.PARENT)
    self.label_posrelparent.set_text('%d, %d' % (bbox.x, bbox.y))
    layer = ci.get_layer()
    self.label_layer.set_text(repr(ci.get_layer()).replace('LAYER_', ''))
    self.label_zorder.set_text(repr(ci.get_mdi_z_order()))
    self.label_alpha.set_text(repr(ci.get_alpha()))
    self.registerEventListener(self._accEventComponent,
                               'object:bounds-changed',
                               'object:visible-data-changed')
  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.label_posrelwin.set_text('')
    self.label_posrelparent.set_text('')
    self.label_posabs.set_text('')
    self.label_size.set_text('')
    self.label_layer.set_text('')
    self.label_zorder.set_text('')
    self.label_alpha.set_text('')

  def _accEventComponent(self, event):
    '''
    Callback for whenever any of the component attributes change.

    @param event: Evnt that triggered this callback.
    @type event: Accessibility.Event
    '''
    if event.source == self.node.acc:
      self.populateUI(event.source)

class _SectionDocument(_InterfaceSection):
  interface_name = 'Document'
  '''
  A class that populates an Component interface section.

  @ivar attr_model: Attribute data model
  @type attr_model: gtk.ListStore
  @ivar label_locale: Locale label
  @type label_locale: gtk.Label
  '''

  def init(self, ui_xml):
    '''
    Initialization that is specific to the Document interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure document attributes tree view
    self.attr_model = ui_xml.get_object('docattrib_liststore')
    self.label_locale = ui_xml.get_object('label_doc_locale')

  def populateUI(self, acc):
    '''
    Populate the Document section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    di = acc.queryDocument()

    self.label_locale.set_text(di.getLocale())

    try:
      attribs = di.getAttributes()
    except:
      attribs = None
    if attribs:
      for attr in attribs:
        name, value = attr.split(':', 1)
        self.attr_model.append([name, value])

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.attr_model.clear()
    self.label_locale.set_text('')

class _SectionCollection(_InterfaceSection):
  interface_name = 'Collection'
  '''
  A placeholder class for Collection interface section.
  '''

class _SectionDesktop(_InterfaceSection):
  interface_name = 'Desktop'
  '''
  A placeholder class for Desktop interface section.
  '''

class _SectionLoginHelper(_InterfaceSection):
  interface_name = 'LoginHelper'
  '''
  A placeholder class for LoginHelper interface section.
  '''

class _SectionHyperlink(_InterfaceSection):
  '''
  A class that populates a Hyperlink interface section.
  '''
  interface_name = 'Hyperlink'

  def init(self, ui_xml):
    '''
    Initialization that is specific to the Hyperlink interface.

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.anchor_count_label = ui_xml.get_object('hyperlink_anchorcount_label')
    self.start_index_label = ui_xml.get_object('hyperlink_startindex_label')
    self.end_index_label = ui_xml.get_object('hyperlink_endindex_label')
    self.valid_label = ui_xml.get_object('hyperlink_valid_label')

    # configure anchors tree view
    treeview = ui_xml.get_object('hyperlink_treeview')
    self.link_model = gtk.TreeStore(int, # index
                          str, # Name of object
                          str, # URI
                          object) # Anchor object
    treeview.set_model(self.link_model)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Name'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 1)
    treeview.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('URI'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 2)
    treeview.append_column(tvc)

    selection = treeview.get_selection()
    show_button = ui_xml.get_object('hyperlink_button_show')
    show_button.set_sensitive(self._isSelectedInView(selection))
    selection.connect('changed', self._onViewSelectionChanged, show_button)

  def populateUI(self, acc):
    '''
    Populate the Hyperlink section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    link = acc.queryHyperlink()

    self.anchor_count_label.set_text(str(link.get_n_anchors()))
    self.start_index_label.set_text(str(link.get_start_index()))
    self.end_index_label.set_text(str(link.get_end_index()))
    self.valid_label.set_text(str(link.is_valid()))

    for anchor_index in range(link.nAnchors):
      acc_obj = link.getObject(anchor_index)
      self.link_model.append(None,
                             [anchor_index, acc_obj.name,
                              link.getURI(anchor_index), acc_obj])

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.link_model.clear()

  def _onShowSelected(self, link_view, *more_args):
    '''
    Callback for row activation or button press. Selects the related
    accessible in the main application.

    @param link_view: The links tree view.
    @type link_view: gtk.TreeView
    @param *more_args: More arguments that are provided by signals,
                       but we discard them all.
    @type *more_args: list
    '''
    selection = link_view.get_selection()
    model, iter = selection.get_selected()
    if iter:
      acc = model[iter][3]
      if acc:
        self.node.update(acc)

class _SectionHypertext(_InterfaceSection):
  '''
  A class that populates an Hypertext interface section.

  @ivar links_model: Data model for available links.
  @type links_model: gtk.ListStore
  '''
  interface_name = 'Hypertext'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Hypertext interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure links tree view
    treeview = ui_xml.get_object('treeview_links')
    # It's a treestore because of potential multiple anchors
    self.links_model = gtk.TreeStore(int, # Link index
                          str, # Name
                          str, # Description
                          str, # URI
                          int, # Start offset
                          int, # End offset
                          object) # Anchor object
    treeview.set_model(self.links_model)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Name'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 1)
    treeview.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('URI'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 3)
    treeview.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Start'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 4)
    treeview.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('End'))
    tvc.set_sizing(gtk.TreeViewColumnSizing.AUTOSIZE)
    tvc.set_resizable(True)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', 5)
    treeview.append_column(tvc)

    selection = treeview.get_selection()
    show_button = ui_xml.get_object('button_hypertext_show')
    show_button.set_sensitive(self._isSelectedInView(selection))
    selection.connect('changed', self._onViewSelectionChanged, show_button)


  def populateUI(self, acc):
    '''
    Populate the Hypertext section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    hti = acc.queryHypertext()

    for link_index in range(hti.getNLinks()):
      link = hti.getLink(link_index)
      iter = self.links_model.append(None,
                                     [link_index,
                                      '', '', '',
                                      link.startIndex,
                                      link.endIndex, None])
      for anchor_index in range(link.nAnchors):
        acc_obj = link.getObject(anchor_index)
        self.links_model.append(iter,
                                [link_index, acc_obj.name, acc_obj.description,
                                 link.getURI(anchor_index),
                                 link.startIndex, link.endIndex, acc_obj])
        if anchor_index == 0:
          self.links_model[iter][1] = \
              acc_obj.name # Otherwise the link is nameless.


  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.links_model.clear()

  def _onLinkShow(self, link_view, *more_args):
    '''
    Callback for row activation or button press. Selects the related
    link accessible in the main application.

    @param link_view: The links tree view.
    @type link_view: gtk.TreeView
    @param *more_args: More arguments that are provided by variuos types of
    signals, but we discard them all.
    @type *more_args: list
    '''
    selection = link_view.get_selection()
    model, iter = selection.get_selected()
    if iter:
      acc = model[iter][6]
      if acc:
        self.node.update(acc)


class _SectionImage(_InterfaceSection):
  '''
  A class that populates an Image interface section.

  @ivar label_pos: Position label
  @type label_pos: gtk.Label
  @ivar label_size: Size label
  @type label_size: gtk.Label
  '''
  interface_name = 'Image'

  def init(self, ui_xml):
    '''
    Initialization that is specific to the Image interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.label_pos = ui_xml.get_object('img_position_label')
    self.label_size = ui_xml.get_object('img_size_label')
    self.label_locale = ui_xml.get_object('img_locale_label')
    self.label_desc = ui_xml.get_object('img_locale_label')

  def populateUI(self, acc):
    '''
    Populate the Image section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ii = acc.queryImage()

    bbox = ii.getImageExtents(pyatspi.DESKTOP_COORDS)
    self.label_pos.set_text('%d, %d' % (bbox.x, bbox.y))
    self.label_size.set_text('%dx%d' % (bbox.width, bbox.height))
    self.label_desc.set_label(ii.imageDescription or \
                                _('(no description)'))
    self.label_locale.set_text(ii.imageLocale)
  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.label_pos.set_text('')
    self.label_size.set_text('')

class _SectionSelection(_InterfaceSection):
  '''
  A class that populates a Selection interface section.

  @ivar sel_model: Data model for child selection options.
  @type sel_model: gtk.ListStore
  @ivar sel_selection: Selection in selection treeview.
  @type sel_selection: gtk.TreeSelection
  @ivar button_select_all: Button for selecting all of the selection nodes.
  @type button_select_all: gtk.Button
  '''
  interface_name = 'Selection'

  def init(self, ui_xml):
    '''
    Initialization that is specific to the Selection interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure selection tree view
    treeview = ui_xml.get_object('treeview_selection')
    self.sel_model = gtk.ListStore(GdkPixbuf.Pixbuf, str, object)
    treeview.set_model(self.sel_model)
    # connect selection changed signal
    self.sel_selection = treeview.get_selection()
    show_button = ui_xml.get_object('button_select_clear')
    show_button.set_sensitive(self._isSelectedInView(self.sel_selection))
    self.sel_selection.connect('changed',
                               self._onViewSelectionChanged,
                               show_button)
    self.sel_selection.connect('changed',
                               self._onSelectionSelected)
    self.button_select_all = ui_xml.get_object('button_select_all')

  def populateUI(self, acc):
    '''
    Populate the Selection section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    if acc.childCount > 50:
      theme = gtk.IconTheme.get_default()
      try:
        warning_icon = theme.load_icon('dialog-warning', 24,
                                       gtk.IconLookupFlags.USE_BUILTIN)
      except:
        warning_icon = None

      self.sel_model.append([warning_icon, _('Too many selectable children'),
                             None])
      # Set section as insensitive, but leave expander label sensitive.
      section_widgets = self.expander.get_children()
      section_widgets.remove(self.expander.get_label_widget())
      for child in section_widgets:
        child.set_sensitive(False)
      return

    for child in acc:
      if child is not None:
        state = child.getState()
        if state.contains(pyatspi.STATE_SELECTABLE):
          self.sel_model.append([getIcon(child), child.name, child])

    state = acc.getState()
    multiple_selections = state.contains(pyatspi.STATE_MULTISELECTABLE)

    self.button_select_all.set_sensitive(multiple_selections)

    if multiple_selections:
      self.sel_selection.set_mode = gtk.SelectionMode.MULTIPLE
    else:
      self.sel_selection.set_mode = gtk.SelectionMode.SINGLE

  def _onSelectionSelected(self, selection):
    '''
    Callback for selection change in the selection treeview. Confusing?

    @param selection: The treeview's selection object.
    @type selection: gtk.TreeSelection
    '''
    try:
      si = self.node.acc.querySelection()
    except NotImplementedError:
      return

    model, paths = selection.get_selected_rows()
    selected_children = [model.get_value(model.get_iter(path), 2).getIndexInParent() for path in paths]

    for child_index in range(len(self.node.acc)):
      if child_index in selected_children:
        si.selectChild(child_index)
      else:
        si.deselectChild(child_index)

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.sel_model.clear()

  def _onSelectionClear(self, widget):
    '''
    Callback for selection clear button.

    @param widget: Widget that triggered callback.
    @type widget: gtk.Widget
    '''
    si = self.node.acc.querySelection()

    self.sel_selection.unselect_all()
    si.clearSelection()

  def _onSelectAll(self, widget):
    '''
    Callback for selection select all button.

    @param widget: Widget that triggered callback.
    @type widget: gtk.Widget
    '''
    si = self.node.acc.querySelection()
    si.selectAll()

class _SectionStreamableContent(_InterfaceSection):
  '''
  A class that populates a StreamableContent interface section.

  @ivar streams_model: Data model for available streams.
  @type streams_model: gtk.ListStore
  '''
  interface_name = 'StreamableContent'

  def init(self, ui_xml):
    '''
    Initialization that is specific to the StreamableContent interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure streamable content tree view
    self.streams_model = ui_xml.get_object('streams_liststore')

  def populateUI(self, acc):
    '''
    Populate the StreamableContent section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    sci = acc.queryStreamableContent()

    for content_type in sci.getContentTypes():
      self.streams_model.append([content_type,
                                 sci.getURI(content_type)])

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.streams_model.clear()

class _SectionTable(_InterfaceSection):
  '''
  A class that populates a Selection interface section.

  @ivar selected_frame: Container frame for selected cell info.
  @type selected_frame: gtk.Frame
  @ivar caption_label: Caption label.
  @type caption_label: gtk.Label
  @ivar summary_label: Summary label.
  @type summary_label: gtk.Label
  @ivar rows_label: Rows label.
  @type rows_label: gtk.Label
  @ivar columns_label: Columns label.
  @type columns_label: gtk.Label
  @ivar srows_label: Selected rows label.
  @type srows_label: gtk.Label
  @ivar scolumns_label: Scolumns label.
  @type scolumns_label: gtk.Label
  @ivar row_ext_label: Row extents label.
  @type row_ext_label: gtk.Label
  @ivar col_ext_label: Column extents label.
  @type col_ext_label: gtk.Label
  @ivar hrow_button: Row header button.
  @type hrow_button: gtk.Button
  @ivar hcol_button: Column header button.
  @type hcol_button: gtk.Button
  @ivar cell_button: Cell button.
  @type cell_button: gtk.Button
  '''
  interface_name = 'Table'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Table interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.selected_frame = ui_xml.get_object('selected_cell_frame')
    self.caption_label = ui_xml.get_object('table_caption_label')
    self.summary_label = ui_xml.get_object('table_summary_label')
    self.rows_label = ui_xml.get_object('table_rows_label')
    self.columns_label = ui_xml.get_object('table_columns_label')
    self.srows_label = ui_xml.get_object('table_srows_label')
    self.scolumns_label = ui_xml.get_object('table_scolumns_label')
    self.row_ext_label = ui_xml.get_object('table_row_extents')
    self.col_ext_label = ui_xml.get_object('table_column_extents')
    self.col_ext_label = ui_xml.get_object('table_column_extents')
    self.hrow_button = ui_xml.get_object('table_hrow_button')
    self.hcol_button = ui_xml.get_object('table_hcol_button')
    self.cell_button = ui_xml.get_object('table_cell_button')

  def populateUI(self, acc):
    '''
    Populate the Table section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ti = acc.queryTable()
    self.selected_frame.set_sensitive(False)
    for attr, label in [(ti.caption, self.caption_label),
                        (ti.summary, self.summary_label),
                        (ti.nRows, self.rows_label),
                        (ti.nColumns, self.columns_label),
                        (ti.nSelectedRows, self.srows_label),
                        (ti.nSelectedColumns, self.scolumns_label)]:
      label.set_text(str(attr))
    self.registerEventListener(self._accEventTable,
                               'object:active-descendant-changed')

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    self.caption_label.set_text('')
    self.summary_label.set_text('')
    self.rows_label.set_text('')
    self.columns_label.set_text('')
    self.srows_label.set_text('')
    self.scolumns_label.set_text('')
    self.row_ext_label.set_text('')
    self.col_ext_label.set_text('')
    self.col_ext_label.set_text('')
    self.hrow_button.set_label('')
    self.hcol_button.set_label('')
    self.cell_button.set_label('')


  def _accEventTable(self, event):
    '''
    Callback for 'object:active-descendant-changed' to detect selected cell
    changes.

    @param event: The event that triggered the callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return

    acc = event.source
    ti = acc.queryTable()
    self.selected_frame.set_sensitive(True)
    is_cell, row, column, rextents, cextents, selected = \
        ti.getRowColumnExtentsAtIndex(event.any_data.getIndexInParent())

    for attr, label in [(rextents, self.row_ext_label),
                             (cextents, self.col_ext_label),
                             (ti.nSelectedRows, self.srows_label),
                             (ti.nSelectedColumns, self.scolumns_label)]:
      label.set_text(str(attr))

    for desc, acc, button in [(ti.getRowDescription(row),
                               ti.getRowHeader(row),
                               self.hrow_button),
                              (ti.getColumnDescription(column),
                               ti.getColumnHeader(column),
                               self.hcol_button),
                              ('%s (%s, %s)' % (event.any_data, row, column),
                               event.any_data,
                               self.cell_button)]:
      button.set_label(str(desc or '<no description>'))
      button.set_sensitive(bool(acc))
      setattr(button, 'acc', acc)

  def _onTableButtonClicked(self, button):
    '''
    Callback for buttons that represent headers.
    Will make the header the main application's selection.

    @param button: Button that triggered event.
    @type button: gtk.Button
    '''
    self.node.update(getattr(button, 'acc'))

class _SectionTableCell(_InterfaceSection):
  '''
  A class that populates a TableCell interface section.
  @ivar row_label: Row label.
  @type row_label: gtk.Label
  @ivar column_label: Column label.
  @type column_label: gtk.Label
  @ivar row_span_label: Row span label.
  @type row_span_label: gtk.Label
  @ivar column_span_label: Column span label.
  @type column_span_label: gtk.Label
  @ivar table_button: Table button.
  @type table_button: gtk.Button
  '''
  interface_name = 'TableCell'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the TableCell interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.row_label = ui_xml.get_object('tablecell_row_label')
    self.column_label = ui_xml.get_object('tablecell_column_label')
    self.row_span_label = ui_xml.get_object('tablecell_rowspan_label')
    self.col_span_label = ui_xml.get_object('tablecell_columnspan_label')
    self.table_button = ui_xml.get_object('tablecell_table_button')


    self.row_headers_model = gtk.ListStore(GdkPixbuf.Pixbuf, str, object)
    self.col_headers_model = gtk.ListStore(GdkPixbuf.Pixbuf, str, object)
    for treeview_id, button_id, model in [('tablecell_rowheaders_treeview',
                                           'tablecell_rowheader_show_button',
                                           self.row_headers_model),
                                          ('tablecell_columnheaders_treeview',
                                           'tablecell_columnheader_show_button',
                                           self.col_headers_model)]:
      treeview = ui_xml.get_object(treeview_id)
      treeview.set_model(model)
      # connect selection changed signal to allow clicking button
      # when a header cell is selected
      selection = treeview.get_selection()
      show_header_button = ui_xml.get_object(button_id)
      show_header_button.set_sensitive(self._isSelectedInView(selection))
      selection.connect('changed', self._onViewSelectionChanged, show_header_button)

  def populateUI(self, acc):
    '''
    Populate the TableCell section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    ti = acc.queryTableCell()
    ok, rowIndex, columnIndex = ti.position
    for attr, label in [(rowIndex, self.row_label),
                        (columnIndex, self.column_label),
                        (ti.rowSpan, self.row_span_label),
                        (ti.columnSpan, self.col_span_label)]:
      label.set_text(str(attr))

    for header_cells, model in [(ti.rowHeaderCells, self.row_headers_model),
                                (ti.columnHeaderCells, self.col_headers_model)]:
      for header in header_cells:
        if header is not None:
          model.append([getIcon(header), header.name, header])

    table = ti.table
    self.table_button.set_label(table.name if (table and table.name) else str(table))
    self.table_button.set_sensitive(bool(table))
    setattr(self.table_button, 'acc', table)

  def clearUI(self):
    '''
    Clear all TableCell specific data.
    '''
    self.row_label.set_text('')
    self.column_label.set_text('')
    self.row_span_label.set_text('')
    self.col_span_label.set_text('')
    self.row_headers_model.clear()
    self.col_headers_model.clear()
    self.table_button.set_label('')

  def _onTableCellButtonClicked(self, button):
    '''
    Callback for button that represents the table.
    Will select the corresponding accessible in the application.

    @param button: Button that triggered event.
    @type button: gtk.Button
    '''
    self.node.update(getattr(button, 'acc'))

  def _onHeaderCellShow(self, header_cells_treeview, *more_args):
    '''
    Callback for "Show" button press. Selects the related
    accessible from the treeview in the main application.

    @param header_cells_treeview: The row/column header cell treeview.
    @type header_cells_treeview: gtk.TreeView
    @param *more_args: More arguments that are provided by various types
    of signals, but we discard them all.
    @type *more_args: list
    '''
    selection = header_cells_treeview.get_selection()
    model, iter = selection.get_selected()
    if iter and model[iter][2]:
      header_cell = model[iter][2]
      self.node.update(header_cell)

class _SectionText(_InterfaceSection):
  '''
  A class that populates a Text interface section.

  @ivar attr_model: Data model for text attributes.
  @type attr_model: gtk.ListStore
  @ivar offset_spin: Offset spinner.
  @type offset_spin: gtk.SpinButton
  @ivar text_view: Text view of provided text.
  @type text_view: gtk.TextView
  @ivar label_caret_offset: Label for caret offset
  @type text_view: gtk.Label
  @ivar text_buffer: Text buffer of provided text.
  @type text_buffer: gtk.TextBuffer
  @ivar toggle_defaults: Toggle button for viewing default text attributes.
  @type toggle_defaults: gtk.CheckButton
  @ivar label_start: Label for current attributes start offset.
  @type label_start: gtk.Label
  @ivar label_end: Label for current attributes end offset.
  @type label_end: gtk.Label
  @ivar _text_insert_handler: Handler ID for text insert events.
  @type _text_insert_handler: integer
  @ivar _text_delete_handler: Handler ID for text delete events.
  @type _text_delete_handler: integer
  @ivar outgoing_calls: Cached outgoing calls to avoid circular event
  invocation.
  @type outgoing_calls: dictionary
  '''
  interface_name = 'Text'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Text interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    # configure text attribute tree view
    self.attr_model = ui_xml.get_object('textattrib_liststore')

    self.offset_spin = ui_xml.get_object('spinbutton_text_offset')
    self.text_view = ui_xml.get_object('textview_text')
    self.label_caret_offset = ui_xml.get_object('label_caret_offset')
    pango_ctx = self.text_view.get_pango_context()
    for f in pango_ctx.list_families():
        name = f.get_name()
        # These are known to show e.g U+FFFC
        if name in [ "Courier New", "Liberation Sans" ]:
            self.text_view.modify_font(Pango.FontDescription(name))
            break

    self.text_buffer = self.text_view.get_buffer()
    self.toggle_defaults = ui_xml.get_object('checkbutton_text_defaults')
    self.label_start = ui_xml.get_object('label_text_attr_start')
    self.label_end = ui_xml.get_object('label_text_attr_end')

    self._text_insert_handler = 0
    self._text_delete_handler = 0

    mark = self.text_buffer.create_mark('attr_mark',
                                        self.text_buffer.get_start_iter(), True)
    self.text_buffer.create_tag('attr_region', foreground='red')

    self.text_buffer.connect('mark-set', self._onTextMarkSet)

    mark.set_visible(True)


    self.text_buffer.connect('modified-changed',
                             self._onTextModified)
    self.text_buffer.connect('notify::cursor-position',
                             self._onTextCursorMove)

    self.text_buffer.set_modified(False)

    # Initialize fifos to help eliminate the viscous cycle of signals.
    # It would be nice if we could just block/unblock it like in gtk, but
    # since it is IPC, asynchronous and not atomic, we are forced to do this.
    self.outgoing_calls = {'itext_insert': self.CallCache(),
                           'itext_delete': self.CallCache()}

  def populateUI(self, acc):
    '''
    Populate the Text section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    self.offset_spin.set_value(0)

    ti = acc.get_text_iface()

    text = Atspi.Text.get_text(ti, 0, ti.get_character_count())
    self.text_buffer.set_text(text)

    self.label_caret_offset.set_text(str(ti.get_caret_offset()))

    self.offset_spin.get_adjustment().upper = ti.get_character_count()

    self.popTextAttr(offset=0)

    eti = acc.get_editable_text_iface()

    expander_label = self.expander.get_label_widget()
    label_text = expander_label.get_label()
    label_text = label_text.replace(_('(Editable)'), '')
    label_text = label_text.strip(' ')
    if eti and acc.get_state_set().contains(Atspi.StateType.EDITABLE):
      label_text += ' ' + _('(Editable)')
      self.text_view.set_editable(True)
    else:
      self.text_view.set_editable(False)
    expander_label.set_label(label_text)

    self._text_insert_handler = self.text_buffer.connect('insert-text',
                                                         self._onITextInsert)
    self._text_delete_handler = self.text_buffer.connect('delete-range',
                                                         self._onITextDelete)

    self.registerEventListener(self._accEventText,
                               'object:text-changed')
    self.registerEventListener(self._accEventTextCaretMoved,
                               'object:text-caret-moved')

  def clearUI(self):
    '''
    Clear all section-specific data.
    '''
    if self._text_insert_handler:
      self.text_buffer.disconnect(self._text_insert_handler)
      self._text_insert_handler = 0
    if self._text_delete_handler:
      self.text_buffer.disconnect(self._text_delete_handler)
      self._text_delete_handler = 0

    self.offset_spin.set_value(0)
    self.label_start.set_text('')
    self.label_end.set_text('')
    self.text_buffer.set_text('')
    self.attr_model.clear()

  def _onTextModified(self, text_buffer):
    '''
    Callback that is triggered when the main text buffer is modified.
    Allows to adjust the spinners bounds accordingly.

    @param text_buffer: The section's main text buffer.
    @type text_buffer: gtk.TextBuffer
    '''
    self.offset_spin.get_adjustment().upper = text_buffer.get_char_count()
    self.offset_spin.set_range(0, text_buffer.get_char_count())
    text_buffer.set_modified(False)

  def _onTextMarkSet(self, text_buffer, iter, text_mark):
    '''
    Callback that is triggered when the attribute mark is moved to
    allow repopulation of the attributes view.

    @param text_buffer: The section's main text buffer.
    @type text_buffer: gtk.TextBuffer
    @param iter: Text iter of the new mark position.
    @type iter: gtk.TextIter
    @param text_mark: The text mark that was moved.
    @type text_mark: gtk.TextMark
    '''
    self.popTextAttr()

  def _onTextSpinnerChanged(self, spinner):
    '''
    Callback for hen the spinner's value changes.
    Moves attribute mark accordingly.

    @param spinner: The marker offset spinner.
    @type spinner: gtk.SpinButton
    '''
    iter = self.text_buffer.get_iter_at_offset(int(self.offset_spin.get_value()))
    self.text_buffer.move_mark_by_name('attr_mark', iter)

  def _onDefaultsToggled(self, toggle_button):
    '''
    Callback for when the "defaults" checkbutton is toggled. Re-populates
    attributes view.

    @param toggle_button: The defaults checkbutton
    @type toggle_button: gtk.CheckButton
    '''
    self.popTextAttr()

  def popTextAttr(self, offset=None):
    '''
    Populate the attributes view with attributes at the given offset, or at
    the attribute mark.

    @param offset: Offset of wanted attributes. If none is given,
    use attribute mark's offset.
    @type offset: integer
    '''
    ti = self.node.acc.get_text_iface()
    if not ti:
      return

    if offset is None:
      mark = self.text_buffer.get_mark('attr_mark')
      iter = self.text_buffer.get_iter_at_mark(mark)
      offset = iter.get_offset()

    show_default = self.toggle_defaults.get_active()
    attr, start, end = ti.get_text_attributes(offset)
    if show_default:
      attr_dict = ti.get_default_attributes()
      attr_dict.update(attr)
    else:
      attr_dict = attr

    attr_list = list(attr_dict.keys())
    attr_list.sort()

    self.attr_model.clear()
    for attr in attr_list:
      self.attr_model.append([attr, attr_dict[attr]])

    self.text_buffer.remove_tag_by_name(
      'attr_region',
      self.text_buffer.get_start_iter(),
      self.text_buffer.get_end_iter())
    self.text_buffer.apply_tag_by_name(
      'attr_region',
      self.text_buffer.get_iter_at_offset(start),
      self.text_buffer.get_iter_at_offset(end))

    # Translators: This string appears in Accerciser's Interface Viewer
    # and refers to a range of characters which has a particular format.
    # "Start" is the character offset where the formatting begins. If
    # the first four letters of some text is bold, the start offset of
    # that bold formatting is 0.
    self.label_start.set_markup(_('Start: %d') % start)
    # Translators: This string appears in Accerciser's Interface Viewer
    # and refers to a range of characters which has a particular format.
    # "End" is the character offset where the formatting ends.  If the
    # first four letters of some text is bold, the end offset of that
    # bold formatting is 4.
    self.label_end.set_markup(_('End: %d') % end)

  def _onTextCursorMove(self, text_buffer, param_spec):
    '''
    Update spinner when input cursor moves.
    Once the spin button's value changes, it's own callback fires
    which re-populates the attribute view.

    @param text_buffer: The section's main text buffer.
    @type text_buffer: gtk.TextBuffer
    @param param_spec: Some gobject crud
    @type param_spec: object
    '''

    self.offset_spin.set_value(text_buffer.get_property('cursor-position'))

    s = text_buffer.get_selection_bounds()
    if s != ():
      # Highlight selected text
      try:
        start,end = s
        startOffset = start.get_offset()
        endOffset = end.get_offset()
        text = self.node.acc.get_text_iface()
        rect = text.get_range_extents(startOffset, endOffset, Atspi.CoordType.SCREEN)
        ah = node._HighLight(rect.x, rect.y, rect.width, rect.height,
                             node.FILL_COLOR, node.FILL_ALPHA,
                             node.BORDER_COLOR, node.BORDER_ALPHA,
                             2.0, 0)
        ah.highlight(node.HL_DURATION)
      except:
        pass

  def _accEventText(self, event):
    '''
    Callback for accessible text changes. Updates the text buffer accordingly.

    @param event: Event that triggered this callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return

    if event.type.major == 'text-changed':
      current_text = self.text_buffer.get_text(self.text_buffer.get_start_iter(),
                                               self.text_buffer.get_end_iter(), False)
      if Atspi.Text.get_text(event.source.get_text_iface(), 0, -1) == current_text:
        # text already up to date
        return

      text_iter = self.text_buffer.get_iter_at_offset(event.detail1)
      if event.type.minor == 'insert':
        call = (event.detail1, event.any_data, event.detail2)
        if self.outgoing_calls['itext_insert'].isCached(call):
          return
        self.text_buffer.handler_block(self._text_insert_handler)
        self.text_buffer.insert(text_iter, event.any_data)
        self.text_buffer.handler_unblock(self._text_insert_handler)

      elif event.type.minor == 'delete':
        call = (event.detail1, event.detail1 + event.detail2)
        if self.outgoing_calls['itext_delete'].isCached(call):
          return
        text_iter_end = \
            self.text_buffer.get_iter_at_offset(event.detail1 + event.detail2)
        self.text_buffer.handler_block(self._text_delete_handler)
        self.text_buffer.delete(text_iter, text_iter_end)
        self.text_buffer.handler_unblock(self._text_delete_handler)

  def _accEventTextCaretMoved(self, event):
    '''
    Callback for object:text-caret-moved events. Updates the displayed
    caret position accordingly.

    @param event: Event that triggered this callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return

    ti = self.node.acc.get_text_iface()
    if not ti:
      return

    self.label_caret_offset.set_text(str(ti.get_caret_offset()))

  def _onITextInsert(self, text_buffer, iter, text, length):
    '''
    Callback for text inserts in the text buffer. Sends changes to examined
    accessible.

    @param text_buffer: The section's main text buffer.
    @type text_buffer: gtk.TextBuffer
    @param iter: Text iter in which the insert occured.
    @type iter: gtk.TextIter
    @param text: The text that was inserted
    @type text: string
    @param length: The length of theinserted text.
    @type length: integer
    '''
    eti = self.node.acc.get_editable_text_iface()
    if not eti:
      return

    call = (iter.get_offset(), text, length)

    self.outgoing_calls['itext_insert'].append(call)
    eti.insert_text(*call)

  def _onITextDelete(self, text_buffer, start, end):
    '''
    Callback for text deletes in the text buffer. Sends changes to examined
    accessible.

    @param text_buffer: The section's main text buffer.
    @type text_buffer: gtk.TextBuffer
    @param start: The start offset of the delete action.
    @type start: integer
    @param end: The end offset of the delete action.
    @type end: integer
    '''
    eti = self.node.acc.get_editable_text_iface()
    if not eti:
      return

    call = (start.get_offset(), end.get_offset())

    self.outgoing_calls['itext_delete'].append(call)
    eti.delete_text(*call)

  def _onTextFocusChanged(self, text_view, event):
    '''
    Callback for leaving and entering focus from the textview,
    it hides/shows the attribute marker. When the textview has focus
    there is no need for the marker to be visible because of th input
    cursor.

    @param text_view: The text view that is being entered or leaved.
    @type text_view: gtk.TextView
    @param event: The focus event.
    @type event: gtk.gdk.Event
    '''
    mark = self.text_buffer.get_mark('attr_mark')
    mark.set_visible(not event.in_)

  class CallCache(list):
    '''
    A list derivative that provides a method for checking if something
    is in the list and removing it at the same time.
    '''
    def isCached(self, obj):
      '''
      Checks if a certain object is in this list instance. If it is, return
      True and remove it.

      @param obj: Object to check for.
      @type obj: object

      @return: True if it is in the list.
      @rtype: boolean
      '''
      if obj in self:
        self.remove(obj)
        return True
      else:
        return False

class _SectionValue(_InterfaceSection):
  '''
  A class that populates a Value interface section.

  @ivar spinbutton: Value spinner.
  @type spinbutton: gtk.SpinButton
  @ivar label_max: Label of maximal value.
  @type label_max: gtk.Label
  @ivar label_min: Label of minimal value.
  @type label_min: gtk.Label
  @ivar label_inc: Label of minimal value increment.
  @type label_inc: gtk.Label
  '''
  interface_name = 'Value'
  def init(self, ui_xml):
    '''
    Initialization that is specific to the Value interface
    (construct data models, connect signals to callbacks, etc.)

    @param ui_xml: Interface viewer glade xml.
    @type ui_xml: gtk.glade.XML
    '''
    self.spinbutton = ui_xml.get_object('spinbutton_value')
    self.label_max = ui_xml.get_object('label_value_max')
    self.label_min = ui_xml.get_object('label_value_min')
    self.label_inc = ui_xml.get_object('label_value_inc')

  def populateUI(self, acc):
    '''
    Populate the Value section with relevant data of the
    currently selected accessible.

    @param acc: The currently selected accessible.
    @type acc: Accessibility.Accessible
    '''
    vi = acc.queryValue()
    self.label_max.set_text(str(vi.maximumValue))
    self.label_min.set_text(str(vi.minimumValue))
    self.label_inc.set_text(str(vi.minimumIncrement))

    minimumIncrement = vi.minimumIncrement
    digits = 0

    while minimumIncrement - int(minimumIncrement) != 0:
      digits += 1
      minimumIncrement *= 10

    # Calling set_range will clamp the value of spinbutton to the allowable
    # range, causing us to try to set the value of the accessible when we
    # really shouldn't.
    self.ignore_value_changes = True
    self.spinbutton.set_range(vi.minimumValue, vi.maximumValue)
    self.ignore_value_changes = False
    self.spinbutton.set_value(vi.currentValue)
    self.spinbutton.set_digits(digits)

    self.registerEventListener(self._accEventValue,
                               'object:property-change:accessible-value',
                               'object:value-changed')

  def _onValueSpinnerChange(self, spinner):
    '''
    Callback for spinner changes. Updates accessible.

    @param spinner: The Value spinner
    @type spinner: gtk.SpinButton
    '''
    if self.ignore_value_changes: return
    vi = self.node.acc.queryValue()
    vi.currentValue = spinner.get_value()

  def _accEventValue(self, event):
    '''
    Callback for value changes from the accessible. Update spin button.

    @param event: The event that triggered the callback.
    @type event: Accessibility.Event
    '''
    if self.node.acc != event.source:
      return
    vi = self.node.acc.queryValue()
    if self.spinbutton.get_value() != vi.currentValue:
      self.spinbutton.set_value(vi.currentValue)



#!/usr/bin/python3

'''
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}


Helper program for WnckWindowManager that retrieves window information
using Wnck and prints them to stdout in JSON format.

The information to print is specified by the program arguments.

This is a separate program to avoid a Wnck (and thus X11 and GTK 3)
dependency in the main application.
'''

import gi
gi.require_version('Wnck', '3.0')
from gi.repository import Wnck

import json
import os
import re
import sys

class WindowInfoProvider:

  def __init__(self):
    # require GDK_BACKEND=x11 environment variable to be set
    # as Wnck only works on X11/XWayland
    if os.environ.get('GDK_BACKEND') != 'x11':
      print('Error: Environment variable GDK_BACKEND=x11 not set', file=sys.stderr)
      exit(1)

    self.wnck_screen = Wnck.Screen.get_default()
    self.wnck_screen.force_update()

  def print_window_infos(self):
    '''
    Print JSON array containing information about all windows to stdout.
    '''
    active_workspace = self.wnck_screen.get_active_workspace()
    win_infos = []
    stacking_index = 0
    for window in self.wnck_screen.get_windows_stacked():
      # client geometry is used unless window has client side decorations,
      # in which case frame geometry is used;
      # assume that client side decoration is used when client rect contains the frame rect
      client_x, client_y, client_width, client_height = window.get_client_window_geometry()
      frame_x, frame_y, frame_width, frame_height = window.get_geometry()
      client_contains_frame = (client_x <= frame_x) and (client_y <= frame_y) \
          and (client_x + client_width >= frame_x + frame_width) \
          and (client_y + client_height >= frame_y + frame_height)
      if client_contains_frame:
        toplevel_x, toplevel_y, toplevel_width, toplevel_height = frame_x, frame_y, frame_width, frame_height
      else:
        toplevel_x, toplevel_y, toplevel_width, toplevel_height = client_x, client_y, client_width, client_height
      title = window.get_name()
      # strip additional trailing Left-to-Right Mark (U+200E), seen with KWin
      if title[-1] == '\u200e':
        title = title[0:-1]
      # KWin 5 (but not KWin 6) adds suffix to window title when there are multiple windows with the
      # same name, e.g. first window: "Hypertext", second window: "Hypertext <2>".
      # Remove the suffix as the accessible name retrieved from AT-SPI2 doesn't have it either
      regex = '^.* <[0-9]+>$'
      if re.match(regex, title):
        title = title[0:title.rfind(' ')]
      # workspace can be None if window is on all workspaces, so only consider case
      # of an actually returned workspace differing from the active one as not on active workspace
      workspace = window.get_workspace()
      on_active_workspace = (not workspace) or (not active_workspace) or window.is_on_workspace(active_workspace)
      pid = window.get_pid()

      win_infos.append(
          {
              'caption': title,
              'geometry.x': toplevel_x,
              'geometry.y': toplevel_y,
              'geometry.width': toplevel_width,
              'geometry.height': toplevel_height,
              'isOnCurrentWorkspace': on_active_workspace,
              'pid': pid,
              'stackingOrder': stacking_index
          }
      )
      stacking_index = stacking_index + 1

    print(json.dumps(win_infos))

  def print_icon_infos(self, window_names):
    '''
    Print relevant information about the window icon that can be
    used to create a GdkPixbuf.Pixbuf from it.
    The information will be printed for the first
    window whose name matches one in the given list.

    @param window_names: List of window names/titles.
    @type list[str]
    '''
    for win in self.wnck_screen.get_windows():
      wname = win.get_name()
      for name in window_names:
        if name == wname:
          pixbuf = win.get_mini_icon()
          result = {
            'pixels': pixbuf.get_pixels().hex(),
            'has_alpha': pixbuf.get_has_alpha(),
            'bits_per_sample': pixbuf.get_bits_per_sample(),
            'width': pixbuf.get_width(),
            'height': pixbuf.get_height(),
            'rowstride': pixbuf.get_rowstride()
          }
          print(json.dumps(result))
          return

    print(json.dumps({}))


if __name__ == '__main__':
  if len(sys.argv) < 2:
    print('Usage: `wnck-get-window-infos.py window-infos` or `wnck-get-window-infos.py icon win_name1 [win_name2 ...]`', file=sys.stderr)
    exit(1)

  provider = WindowInfoProvider()
  if sys.argv[1] == "window-infos":
    provider.print_window_infos()
  elif sys.argv[1] == "icon":
    provider.print_icon_infos(sys.argv[2:])
  else:
    print('Usage: `wnck-get-window-infos.py window-infos` or `wnck-get-window-infos.py icon win_name1 [win_name2 ...]`', file=sys.stderr)



'''
Provides support for internationalization using the Python gettext
module. Configures gettext to use the translation associated with the locale
set by the user's environment at runtime. If no appropriate translation is
found, the default language is used: the one in which the strings in the
code are written (i.e. English).

The gettext function in the gettext module is alised as _ to match the name of
the equivalent C function provided by GNU gettext tools.

See the Python documentation on the gettext module at
U{http://docs.python.org/lib/module-gettext.html}. A brief tutorial on
i18n and Python is available at
U{http://www.async.com.br/faq/pygtk/index.py?req=show&file=faq22.002.htp}

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2005, 2007 IBM Corporation
@license: The BSD License

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD license which accompanies
this distribution, and is available at
U{http://www.opensource.org/licenses/bsd-license.php}
'''
import sys, locale, gettext, os.path
from gettext import gettext as translate_func

def bind(domain, locale_dir):
  '''
  Convenience function for creating a new instance of a Python translation
  class bound to a domain and locale directory other than the default one for
  LSR. This function is useful in L{UIElement}s which have translatable strings
  and ship separately from LSR with their own translation files. For instance,
  a L{Perk} writer may call this function globally in his L{Perk} and assign
  the result to a global variable _ to use the GNU translation system without
  affecting how other components of LSR are translated.

  @param domain: Filename of the translation file, typically the name of the
    project with which it is associated
  @type domain: string
  @param locale_dir: Directory to search for translation files for the domain
  @type locale_dir: string
  @return: Bound method gettext on the translation object
  @rtype: callable
  '''
  try:
    t = gettext.translation(domain, locale_dir)
    return t.gettext
  except IOError:
    # no translation file, so just return the unicode version of the given
    # string
    return lambda x: str(x)

# use the current system locale
locale.setlocale(locale.LC_ALL, '')
# build the path to where the main translation file lives
LOCALE_DIR = '/usr/share/locale'
DOMAIN = 'accerciser'
# build a default instance to the LSR domain and locale directory
_ = bind(DOMAIN, LOCALE_DIR)
gettext.bindtextdomain (DOMAIN, LOCALE_DIR)
gettext.textdomain(DOMAIN)

def C_(ctx, s):
  '''Provide qualified translatable strings.  Some strings translate to
  more than one string in another locale.  We provide a convention to
  provide contextual information for the string so that translators can
  do the right thing.'''
  translated = translate_func('%s\x04%s' % (ctx, s))
  if '\x04' in translated:
    # no translation found, return input string
    return s
  return translated

# allow gettext to extract the string, but it should not be translated inline.
N_ = lambda string: string



'''
Defines and manages the multiple plugin views.

@author: Eitan Isaacson
@organization: Mozilla Foundation
@copyright: Copyright (c) 2006, 2007 Mozilla Foundation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
from gi.repository import Gio as gio
from gi.repository import GLib as glib
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository.Gio import Settings as GSettings
from gi.repository import GObject

from .base_plugin import Plugin
from accerciser.tools import *
from .message import MessageManager
import os
import sys
from accerciser.i18n import _, N_
import gc
from accerciser import menus


GSCHEMA = 'org.a11y.Accerciser'
PLUGVIEWS_GSCHEMA = 'org.a11y.Accerciser.pluginviews'
NEWPLUGVIEWS_GSCHEMA = 'org.a11y.Accerciser.newpluginviews'
NEWPLUGVIEWS_PATH = '/org/a11y/accerciser/newpluginviews/'


class PluginView(gtk.Notebook):
  '''
  Container for multiple plugins. Implemented with a gtk notebook.

  @cvar TARGET_PLUGINVIEW: Drag and drop target ID for another pluginview.
  @type TARGET_PLUGINVIEW: integer
  @cvar TARGET_ROOTWIN: Drag and drop target ID for root window.
  @type TARGET_ROOTWIN: integer
  @ivar NOTEBOOK_GROUP: Group ID for detachable tabs.
  @type NOTEBOOK_GROUP: string

  @ivar view_name: Name of view.
  @type view_name: string
  '''

  __gsignals__ = {'plugin_drag_end' :
                  (GObject.SignalFlags.RUN_FIRST,
                   None,
                   (GObject.TYPE_OBJECT,)),
                  'tab_popup_menu' :
                  (GObject.SignalFlags.RUN_FIRST,
                   None,
                   (GObject.TYPE_PYOBJECT,
                    GObject.TYPE_OBJECT))}
  TARGET_PLUGINVIEW = 0
  TARGET_ROOTWIN = 1
  NOTEBOOK_GROUP = 'NOTEBOOK_GROUP'

  def __init__(self, view_name):
    '''
    Initialize a new plugin view.

    @param view_name: The name of the view.
    @type view_name: string
    '''
    gtk.Notebook.__init__(self)
    self.view_name = view_name
    self.set_scrollable(True)
    self.set_group_name(self.NOTEBOOK_GROUP)
    self.connect('drag-end', self._onDragEnd)
    self.connect('drag-data-get', self._onDragDataGet)
    self.connect('key-press-event', self._onKeyPress)
    self.connect('button-press-event', self._onButtonPress)

    self.dest_type = None

  def _onButtonPress(self, nb, event):
    '''
    Callback for button presses, used for tab context menus.

    @param nb: Notebook that emitted signal.
    @type nb: gtk.Notebook
    @param event: Event object.
    @type event: gtk.dk.Event
    '''
    plugin = self._getClickedPlugin(event.x_root, event.y_root)
    if plugin and event.button == 3:
      self.emit('tab_popup_menu', event, plugin)

  def _onKeyPress(self, nb, event):
    '''
    Callback for key presses, used for tab context menus.

    @param nb: Notebook that emitted signal.
    @type nb: gtk.Notebook
    @param event: Event object.
    @type event: gtk.dk.Event
    '''
    if event.keyval == gdk.KEY_Menu and \
          self.get_property('has-focus'):
      page_num = self.get_current_page()
      child = self.get_nth_page(page_num)
      if isinstance(child, Plugin):
        self.emit('tab_popup_menu', event, self.get_nth_page(page_num))

  def _getClickedPlugin(self, event_x, event_y):
    '''
    Determines which plugin's tab was clicked with given coordinates.

    @param event_x: X coordnate of click.
    @type event_x: integer
    @param event_y: Y coordnate of click.
    @type event_y: integer

    @return: Tab's plugin or None if nothing found.
    @rtype: L{Plugin}
    '''
    for child in self.getPlugins():
      tab = self.get_tab_label(child)
      # TODO-JH: Don't really know if this is correct
      if tab != None and tab.get_state_flags() & tab.get_mapped():
        x, y, w, h = self.getTabAlloc(tab)
        if event_x >= x and \
              event_x <= x + w and \
              event_y >= y and \
              event_y <= y + h:
          return child
    return None

  def getTabAlloc(self, widget):
    '''
    Get the screen allocation of the given tab.

    @param widget: The tab widget.
    @type widget: gtk.Widget

    @return: X, Y, width any height coordinates.
    @rtype: tuple
    '''
    gdk_window = widget.get_window()
    origin_z, origin_x, origin_y = gdk_window.get_origin()
    alloc = widget.get_allocation()
    x, y, width, height = \
        alloc.x, alloc.y, alloc.width, alloc.height
    # TODO-JH: Review this conversion
    has_window = not(widget.get_has_window())
    if bool(widget.get_state_flags().value_names) & has_window:
      origin_x += x
      origin_y += y
    return origin_x, origin_y, width, height

  def _onDragDataGet(self, widget, context, selection_data, info, time):
    '''
    Data transfer function for drag and drop.

    @param widget: Widget that recieved the signal.
    @type widget: gtk.Widget
    @param context: Drag context for this operation
    @type context: gtk.gdk.DragContext
    @param selection_data: A selection data object
    @type selection_data: gtk.SelectionData
    @param info: An ID of the drag
    @type info: integer
    @param time: The timestamp of the drag event.
    @type time: float
    '''
    self.dest_type = info
    selection_data.set(selection_data.get_target(), 8, '')

  def _onDragEnd(self, widget, drag_context):
    '''
    Callback for completed drag operation.

    @param widget: Widget that recieved the signal.
    @type widget: gtk.Widget
    @param drag_context: Drag context for this operation
    @type drag_context: gtk.gdk.DragContext
    '''
    if self.dest_type == self.TARGET_PLUGINVIEW:
      return
    index = self.get_current_page()
    child = self.get_nth_page(index)
    self.emit('plugin_drag_end', child)

  def getPlugins(self):
    '''
    Return list of plugins in given view. Filter out tabs that are not plugins.

    @return: Plugins in given view.
    @rtype: List of {Plugin}
    '''
    return [x for x in self.get_children() if isinstance(x, Plugin)]

  def insert_page(self, child, tab_label=None, position=-1):
    '''
    Override gtk.Notebook's method. Use the plugin's name or widget's name
    as default tab label. Keep message tab as first tab.

    @param child: Child widget to insert.
    @type child: gtk.Widget
    @param tab_label: Label to use. Plugin name will be used by default.
    @type tab_label: gtk.Widget
    @param position: Position to insert child.
    @type position: integer
    '''
    if position != -1 and  \
          isinstance(self.get_nth_page(0), MessageManager.MessageTab):
      position += 1
    if tab_label:
      name = tab_label
    elif isinstance(child, Plugin):
      name = getattr(child, 'plugin_name_localized', None) or child.plugin_name
    elif child.get_name():
      name = child.get_name()
    gtk.Notebook.append_page(self, child, None)
    gtk.Notebook.reorder_child(self, child, position)
    gtk.Notebook.set_tab_label(self, child, gtk.Label.new(name))

  def append_page(self, child, tab_label=None):
    '''
    Override gtk.Notebook's method. Use the plugin's name or widget's name
    as default tab label. Keep message tab as first tab.

    @param child: Child widget to insert.
    @type child: gtk.Widget
    @param tab_label: Label to use. Plugin name will be used by default.
    @type tab_label: gtk.Widget
    '''
    self.insert_page(child, tab_label, -1)

  def prepend_page(self, child, tab_label=None):
    '''
    Override gtk.Notebook's method. Use the plugin's name or widget's name
    as default tab label. Keep message tab as first tab.

    @param child: Child widget to insert.
    @type child: gtk.Widget
    @param tab_label: Label to use. Plugin name will be used by default.
    @type tab_label: gtk.Widget
    '''
    self.insert_page(child, tab_label, 0)

  def focusTab(self, tab_num):
    '''
    Set focus on given tab number.

    @param tab_num: Index within visible tabs.
    @type tab_num: integer
    '''
    children = self.get_children()
    shown_children = [x for x in children if x.get_property('visible')]
    try:
      child = shown_children[tab_num]
    except IndexError:
      return
    self.set_current_page(children.index(child))
    self.grab_focus()

  def getNVisiblePages(self):
    '''
    Get number of visible children.

    @return: Number of visible children.
    @rtype: integer
    '''
    shown_children = [x for x in self.get_children() if x.get_property('visible')]
    return len(shown_children)

class PluginViewWindow(gtk.Window, ToolsAccessor):
  '''
  Standalone window with a plugin view.

  @ivar plugin_view: Embedded plugin view.
  @type plugin_view: L{PluginView}
  '''
  def __init__(self, view_name):
    '''
    Initialize a new plugin view window.

    @param view_name: The name of the view.
    @type view_name: string
    '''
    gtk.Window.__init__(self)
    self.plugin_view = PluginView(view_name)
    self.add(self.plugin_view)

    gspath = NEWPLUGVIEWS_PATH + view_name.lower().replace(' ', '-') + '/'
    gsettings = GSettings.new_with_path(NEWPLUGVIEWS_GSCHEMA, gspath)
    width = gsettings.get_int('width')
    height = gsettings.get_int('height')
    self.set_default_size(width, height)
    self.connect('key_press_event', self._onKeyPress)
    self.plugin_view.connect_after('page_removed', self._onPluginRemoved)
    self.set_title(view_name)
    self.set_position(gtk.WindowPosition.MOUSE)
    self.show_all()
    self.connect('size-allocate', self._onResize)

  def _onResize(self, widget, allocation):
    '''
    Callback for window resizing. Used for persisting view sizes across
    sessions.

    @param widget: Window widget.
    @type widget: gtk.Widget
    @param allocation: The new allocation.
    @type allocation: gtk.gdk.Rectangle
    '''
    view_name = self.plugin_view.view_name
    gspath = NEWPLUGVIEWS_PATH + view_name.lower().replace(' ', '-') + '/'
    gsettings = GSettings.new_with_path(NEWPLUGVIEWS_GSCHEMA, gspath)
    gsettings.set_int('width', self.get_allocated_width())
    gsettings.set_int('height', self.get_allocated_height())

  def _onPluginRemoved(self, pluginview, page, page_num):
    '''
    Callback for removed tabs. If there are no plugins in a stand alone view,
    destroy it.

    @param pluginview: Plugin view that emitted signal.
    @type pluginview: L{PluginView}
    @param page: Child that has been removed.
    @type page: gtk.Widget
    @param page_num: Tab index of child.
    @type page_num: integer
    '''
    if pluginview.get_n_pages() == 0:
      self.destroy()

  def _onKeyPress(self, widget, event):
    '''
    Callback for keypresses in window. Enables alt-<num> tab switching.

    @param widget: Window widget.
    @type widget: gtk.Widget
    @param event: Event object
    @type event: gtk.gdk.Event
    '''
    if event.state & gdk.ModifierType.MOD1_MASK and \
          event.keyval in range(gdk.keyval_from_name('0'),
                                 gdk.keyval_from_name('9')):
      tab_num = event.keyval - gdk.keyval_from_name('0') or 10
      pages_count = self.plugin_view.get_n_pages()
      if pages_count >= tab_num:
        self.plugin_view.focusTab(tab_num - 1)


class ViewManager(object):
  '''
  Manage plugins and their views.
  '''
  def __init__(self, application, *perm_views):
    '''
    Initialize view manager.

    @param application: The application.
    @type application: gtk.Application.
    @param perm_views: List of permanent views, at least one is required.
    @type perm_views: list of {PluginView}
    '''
    self.application = application
    self._perm_views = perm_views
    gsettings = GSettings.new(PLUGVIEWS_GSCHEMA)
    single = gsettings.get_boolean('layout-single')
    self._initViewModel(single)
    self._setupActions()

  def _setupActions(self):
    '''
    Sets up actions related to plugin layout.
    '''
    single = isinstance(self._view_model, SingleViewModel)
    menu, name, label, accel, callback = (menus.view_menu_general_section, 'single_view_mode', \
                                         _('_Single plugins view'), '<Control>t', self._onSingleViewToggled)
    action_name = 'app.' + name
    menu_item = gio.MenuItem.new(label, action_name)
    menu.append_item(menu_item)
    action = gio.SimpleAction.new_stateful(name, None, glib.Variant.new_boolean(single))
    action.connect('change-state', callback)
    self.application.set_accels_for_action(action_name, [accel])
    self.application.add_action(action)

    self.setSingleMode(action.get_property('state').get_boolean())


  def _onSingleViewToggled(self, action, value, data=None):
    '''
    Callback for single view toggle action.

    @param action: Action object that emitted callback.
    @type action: gio.SimpleAction
    '''
    action.set_state(value)
    self.setSingleMode(value.get_boolean())

  def setSingleMode(self, single):
    '''
    Toggle single mode on or off.

    @param single: True if we want single mode.
    @type single: boolean
    '''
    if isinstance(self._view_model, SingleViewModel) == single:
      return
    gsettings = GSettings.new(PLUGVIEWS_GSCHEMA)
    gsettings.set_boolean('layout-single', single)
    plugins = self._view_model.getViewedPlugins()
    self._view_model.close()
    del self._view_model
    for plugin in plugins:
      if plugin.get_parent():
        plugin.get_parent().remove(plugin)
    self._initViewModel(single)
    for plugin in plugins:
      self._view_model.addElement(plugin)

  def _initViewModel(self, single):
    '''
    Initialize a view model, either multi view or single view.

    @param single: True if we want single mode.
    @type single: boolean
    '''
    if single:
      self._view_model = SingleViewModel(*self._perm_views)
    else:
      self._view_model = MultiViewModel(*self._perm_views)

  def addElement(self, element):
    '''
    Add an element to a plugin view.

    @param element: The element to be added to a view.
    @type element: gtk.Widget
    '''
    self._view_model.addElement(element)

  def close(self):
    '''
    Do cleanup.
    '''
    self._view_model.close()

  def initialView(self):
    '''
    Do something when view/s are first displayed.
    '''
    self._view_model.initialView()

  def giveElementFocus(self, element):
    '''
    Give focus to given element (ie. a plugin)

    @param element: The element to give focus to.
    @type element: gtk.Widget
    '''
    self._view_model.giveElementFocus(element)

  def changeView(self, plugin, new_view_name):
    '''
    Put a plugin instance in a different view.

    @param plugin: Plugin to move.
    @type plugin: L{Plugin}
    @param new_view_name: New view name.
    @type new_view_name: string
    '''
    self._view_model.changeView(plugin, new_view_name)

  def getViewNameForPlugin(self, plugin_name):
    '''
    Get the view name for a given plugin name.

    @param plugin_name: Plugin's name to lookup view for.
    @type plugin_name: string

    @return: View name for plugin.
    @rtype: string
    '''
    return self._view_model.getViewNameForPlugin(plugin_name)

  def Menu(self, context_plugin, transient_window):
    '''
    Return a context menu for the given plugin with view manipulation options.

    @param context_plugin: Subject plugin of this menu.
    @type context_plugin: L{Plugin}
    @param transient_window: Transient parent window. Used for keeping the
    new view dialog modal.
    @type transient_window: gtk.Window

    @return: A Menu widget
    @rtype: gtk.Menu
    '''
    return self._view_model.Menu(context_plugin, transient_window)

class BaseViewModel(ToolsAccessor):
  '''
  Base class for views model

  @ivar perm_views: List of permanent views.
  @type perm_views: list of L{PluginView}
  @ivar main_view: Main view.
  @type main_view: L{PluginView}
  '''
  def __init__(self, *perm_views):
    '''
    Initialize view model.

    @param perm_views: List of permanent views, at least one is required.
    @type perm_views: list of {PluginView}
    '''
    if len(perm_views) == 0:
      raise TypeError('View model needs at least one permanent view')
    self.perm_views = perm_views
    self.main_view = perm_views[0]

  def addElement(self, element):
    '''
    Add an element to a plugin view. If the element is a message tab, put it as
    the first tab in the main view. If the element is a plugin, check if it's
    placement is cached in this instance or read it's position from gsettings.
    By default a plugin is appended to the main view.

    @param element: The element to be added to a view.
    @type element: gtk.Widget
    '''
    if isinstance(element, Plugin):
      self.addPlugin(element)
    elif isinstance(element, MessageManager.MessageTab):
      element.connect('show', Proxy(self._onMessageTabShow))
      element.hide()
      self.main_view.insert_page(element, 0)
      self.main_view.set_tab_detachable(element, False)
      self.main_view.set_tab_reorderable(element, False)

  def addPlugin(self, plugin):
    '''
    Add a plugin to the view. Check if it's placement is cached in this
    instance or read it's position from gsettings. By default a plugin is
    appended to the main view.

    @param plugin: Plugin to add.
    @type plugin: L{Plugin}
    '''
    pass

  def close(self):
    '''
    Do cleanup.
    '''
    pass

  def initialView(self):
    '''
    Do something when view/s are first displayed.
    '''
    pass

  def giveElementFocus(self, element):
    '''
    Give focus to given element (ie. a plugin)

    @param element: The element to give focus to.
    @type element: gtk.Widget
    '''
    if not getattr(element, 'parent', None):
      return
    view = element.parent
    page_num = view.page_num(element)
    view.set_current_page(page_num)
    view.set_focus_child(element)

  def _onMessageTabShow(self, message_tab):
    '''
    Callback for when a message tab appears. Give it focus.

    @param message_tab: Message tab that just appeared.
    @type message_tab: L{MessageManager.MessageTab}
    '''
    self.giveElementFocus(message_tab)

  def changeView(self, plugin, new_view_name):
    '''
    Put a plugin instance in a different view.

    @param plugin: Plugin to move.
    @type plugin: L{Plugin}
    @param new_view_name: New view name.
    @type new_view_name: string
    '''
    pass

  def getViewNameForPlugin(self, plugin_name):
    '''
    Get the view name for a given plugin name.

    @param plugin_name: Plugin's name to lookup view for.
    @type plugin_name: string

    @return: View name for plugin.
    @rtype: string
    '''
    pass

  def getViewedPlugins(self):
    '''
    Get a list of all viewed plugins that are in the managed view/s.
    '''
    pass

  def Menu(self, context_plugin, transient_window):
    '''
    Return a context menu for the given plugin with view manipulation options.

    @param context_plugin: Subject plugin of this menu.
    @type context_plugin: L{Plugin}
    @param transient_window: Transient parent window. Used for keeping the
    new view dialog modal.
    @type transient_window: gtk.Window

    @return: A Menu widget
    @rtype: gtk.Menu
    '''
    return gtk.Menu()

class SingleViewModel(BaseViewModel):
  def addPlugin(self, plugin):
    '''
    Add a given plugin to our single view in alphabetical order.

    @param plugin: Plugin to add
    @type plugin: L{Plugin}
    '''
    plugin_names = \
        [p.plugin_name.lower() for p in self.main_view.getPlugins()]
    plugin_names.append(plugin.plugin_name.lower())
    plugin_names.sort()
    index = plugin_names.index(plugin.plugin_name.lower())
    self.main_view.insert_page(plugin, position=index)
    self.main_view.set_tab_detachable(plugin, False)
    self.main_view.set_tab_reorderable(plugin, False)
    plugin.show_all()
  def initialView(self):
    '''
    Set tab to first tab.
    '''
    self.main_view.set_current_page(0)
  def getViewedPlugins(self):
    '''
    Get all managed plugins.

    @return: list of all managed plugins.
    @rtype: list of PLugin
    '''
    return self.main_view.getPlugins()

class MultiViewModel(list, BaseViewModel):
  '''
  Manages all plugin views. Implements a gtk.ListStore of all views.
  Persists plugin view placements across sessions.

  @cvar COL_NAME: View name column ID.
  @type COL_NAME: integer
  @cvar COL_INSTANCE: View instance column ID.
  @type COL_INSTANCE: integer

  @ivar perm_views: List of permanent views.
  @type perm_views: list of L{PluginView}
  @ivar main_view: Main view.
  @type main_view: L{PluginView}
  @ivar _ignore_insertion: A list of tuples with view and plugin names that
  should be ignored and not go in to gsettings. This is to avoid recursive
  gsettings modification.
  @type _ignore_insertion: list of tuples
  @ivar _placement_cache: A cache of recently disabled plugins with their
  placement. allowsthem to be enabled in to the same position.
  @type _placement_cache: dictionary
  @ivar _closed: Indicator to stop writing plugin remove events to gsettings.
  @type _closed: boolean
  '''
  COL_NAME = 0
  COL_INSTANCE = 1
  def __init__(self, *perm_views):
    '''
    Initialize view manager.

    @param perm_views: List of permanent views, at least one is required.
    @type perm_views: list of {PluginView}
    '''
    BaseViewModel.__init__(self, *perm_views)
    for view in self.perm_views:
      self.append(view)
      self._connectSignals(view)
    self._ignore_insertion = []
    self._placement_cache = {}
    self._closed = False

  def close(self):
    '''
    Stops gsettings maniputaion.
    '''
    self._closed = True

  def getViewNameForPlugin(self, plugin_name):
    '''
    Get the view name for a given plugin name as defined in gsettings.
    Or return name of main view.

    @param plugin_name: Plugin's name to lookup view for.
    @type plugin_name: string

    @return: View name for plugin.
    @rtype: string
    '''
    plugin_layouts = self._getPluginLayouts()
    for view_name in plugin_layouts:
      if plugin_name in plugin_layouts[view_name]:
        return view_name
    return self.main_view.view_name

  def _getViewByName(self, view_name):
    '''
    Return the view instance of the given name.

    @param view_name: Name of view to retrieve.
    @type view_name: string

    @return: View instance or None
    @rtype: L{PluginView}
    '''
    for view in self:
      if view.view_name == view_name:
        return view
    return None

  def _onPluginDragEnd(self, view, plugin):
    '''
    Callback for the end of a drag operation of a plugin. Only is called
    when the drag ends on the root window.

    @param view: Current plugin's view.
    @type view: L{PluginView}
    @param plugin: Plugin that was dragged.
    @type plugin: L{Plugin}
    '''
    new_view = self._newView()
    view.remove(plugin)
    new_view.append_page(plugin)
    new_view.set_tab_detachable(plugin, True)
    new_view.set_tab_reorderable(plugin, True)

  def _newView(self, view_name=None):
    '''
    Creates a new view.

    @param view_name: An optional view name. Gives a more mundane one if no
    name is provided.
    @type view_name: string

    @return: New view
    @rtype: L{PluginView}
    '''
    if not view_name:
      view_name = _('Plugin View')
      view_num = 2
      while view_name in self._getViewNames():
        view_name = _('Plugin View (%d)') % view_num
        view_num += 1
    w = PluginViewWindow(view_name)
    view = w.plugin_view
    self._connectSignals(view)
    self.append(view)
    return view

  def _getViewOrNewView(self, view_name):
    '''
    Get an existing or new view with the current name.

    @param view_name: View's name
    @type view_name: string

    @return: New or existing view.
    @rtype: L{PluginView}
    '''
    view = self._getViewByName(view_name) or self._newView(view_name)
    return view

  def _onViewDelete(self, view_window, event):
    '''
    Callback for a view window's delete event. Puts all orphaned plugins
    in main view.

    @param view_window: View window that emitted delete event.
    @type view_window: L{PluginViewWindow}
    @param event: Event object.
    @type event: gtk.gdk.Event
    '''
    view = view_window.plugin_view
    for child in view.getPlugins():
      view.remove(child)
      self.main_view.append_page(child)
    self._removeView(view)

  def _removeView(self, view):
    '''
    Removes view from model.

    @param view: View to remove.
    @type view: L{PluginView}
    '''
    if view in self.perm_views:
      return
    if view in self:
      self.remove(view)

  def _onTabPopupMenu(self, view, event, plugin):
    '''
    Callback for popup menu signal from plugin view. Displays a context menu
    with available views.

    @param view: Plugin view that emitted this signal.
    @type view: L{PluginView}
    @param event: Relevant event object that will be used in popup menu.
    @type event: gtk.gdk.Event
    @param plugin: Plugin of tab that was clicked or pressed.
    @type plugin: L{Plugin}
    '''
    menu = self.Menu(plugin, view.get_toplevel())
    if hasattr(event, 'button'):
      menu.popup(None, None, None, None, event.button, event.time)
    else:
      tab = view.get_tab_label(plugin)
      x, y, w, h = view.getTabAlloc(tab)
      rect = gdk.Rectangle(x, y, w, h)
      menu.popup(None, None,
                 lambda m, r: (r.x+r.width/2, r.y+r.height/2, True),
                 rect, 0, event.time)

  def _connectSignals(self, view):
    '''
    Convenience function for connecting all needed signal callbacks.

    @param view: Plugin view to connect.
    @type view: :{PluginView}
    '''
    if isinstance(view.get_parent(), PluginViewWindow):
      view.get_parent().connect('delete-event', Proxy(self._onViewDelete))
    view.connect('plugin-drag-end', Proxy(self._onPluginDragEnd))
    view.connect('tab-popup-menu', Proxy(self._onTabPopupMenu))
    view.connect('page-added', Proxy(self._onViewLayoutChanged), 'added')
    view.connect('page-removed', Proxy(self._onViewLayoutChanged), 'removed')
    view.connect('page-reordered', Proxy(self._onViewLayoutChanged), 'reordered')

  def _onViewLayoutChanged(self, view, plugin, page_num, action):
    '''
    Callback for all layout changes. Updates gsettings.

    @param view: View that emitted the signal.
    @type view: L{PluginView}
    @param plugin: Plugin that moved.
    @type plugin: L{Plugin}
    @param page_num: Plugin's position in view.
    @type page_num: integer
    @param action: Action that triggered this event.
    @type action: string
    '''
    if self._closed or not isinstance(plugin, Plugin): return
    if (view.view_name, plugin.plugin_name) in self._ignore_insertion:
      self._ignore_insertion.remove((view.view_name, plugin.plugin_name))
      return
    if plugin.plugin_name in self._placement_cache:
      self._placement_cache.pop(plugin.plugin_name)

    plugin_layouts = self._getPluginLayouts()
    try:
      plugin_layout = plugin_layouts[view.view_name]
    except KeyError:
      plugin_layouts[view.view_name] = []
      plugin_layout = plugin_layouts[view.view_name]
    if plugin.plugin_name in plugin_layout:
        plugin_layout.remove(plugin.plugin_name)
    if action in ('reordered', 'added'):
      plugin_layout.insert(page_num, plugin.plugin_name)
    elif action == 'removed':
      self._placement_cache[plugin.plugin_name] = (view.view_name, page_num)
    if len(plugin_layout) == 0:
      self._removeView(view)

    self._setPluginLayouts(plugin_layouts)

  def _setPluginLayouts(self, plugin_layouts):
    self.plugviews = GSettings.new(PLUGVIEWS_GSCHEMA)
    self.plugviews.set_strv('top-panel-layout', plugin_layouts.pop('Top panel'))
    self.plugviews.set_strv('bottom-panel-layout', plugin_layouts.pop('Bottom panel'))

    for plugview in list(plugin_layouts.keys()):
      gspath = NEWPLUGVIEWS_PATH + plugview.lower().replace(' ', '-') + '/'
      newview = GSettings.new_with_path(NEWPLUGVIEWS_GSCHEMA, gspath)
      newview.set_strv('layout', plugin_layouts[plugview])
      l = self.plugviews.get_strv('available-newviews')
      l.append(plugview)
      self.plugviews.set_strv('available-newviews', l)

  def _getPluginLayouts(self):
    plugin_layouts= {}
    self.plugviews = GSettings.new(PLUGVIEWS_GSCHEMA)
    plugin_layouts['Top panel'] = self.plugviews.get_strv('top-panel-layout')
    plugin_layouts['Bottom panel'] = self.plugviews.get_strv('bottom-panel-layout')

    for plugview in self.plugviews.get_strv('available-newviews'):
      gspath = NEWPLUGVIEWS_PATH + plugview.lower().replace(' ', '-') + '/'
      newview = GSettings.new_with_path(NEWPLUGVIEWS_GSCHEMA, gspath)
      layout = newview.get_strv('layout')
      if layout:
        plugin_layouts[plugview] = layout
      else:
        l = self.plugviews.get_strv('available-newviews')
        l.remove(plugview)
        self.plugviews.set_strv('available-newviews', l)
    return plugin_layouts


  def addPlugin(self, plugin):
    '''
    Add a plugin to the view. Check if it's placement is cached in this
    instance or read it's position from gsettings. By default a plugin is
    appended to the main view.

    @param plugin: Plugin to add.
    @type plugin: L{Plugin}
    '''
    if plugin.plugin_name in self._placement_cache:
      view_name, index = self._placement_cache.pop(plugin.plugin_name)
      view = self._getViewOrNewView(view_name)
    else:
      view_name = self.getViewNameForPlugin(plugin.plugin_name)
      view = self._getViewOrNewView(view_name)
      plugin_layouts = self._getPluginLayouts()
      try:
        plugin_layout = plugin_layouts[view.view_name]
      except KeyError:
        plugin_layout = []
        plugin_layouts[view.view_name] = plugin_layout
      index = -1
      if plugin.plugin_name in plugin_layout:
        # The plugins that have a higher index.
        successive = plugin_layout[plugin_layout.index(plugin.plugin_name)+1:]
        for child_index, preceding_plugin in enumerate(view.getPlugins()):
          if preceding_plugin.plugin_name in successive:
            # Place new plugin just before the first successive plugin.
            index = child_index
            break
      self._ignore_insertion.append((view.view_name, plugin.plugin_name))
      self._setPluginLayouts(plugin_layouts)

    view.insert_page(plugin, position=index)
    view.set_tab_detachable(plugin, True)
    view.set_tab_reorderable(plugin, True)
    plugin.show_all()

  def initialView(self):
    '''
    Set the current tab of all views to be the first one.
    Used when Accercier first starts.
    '''
    for view in self:
      view.set_current_page(0)

  def getViewedPlugins(self):
    '''
    Get all plugins from all views.
    '''
    rv = []
    for view in self:
      rv.extend(view.getPlugins())
    return rv

  def _getViewNames(self):
    '''
    Get a list of all managed view names.

    @return: A list of view names.
    @rtype: list of string
    '''
    return [view.view_name for view in self]

  def changeView(self, plugin, new_view_name):
    '''
    Put a plugin instance in a different view. If given view name does not
    exist, create it.

    @param plugin: Plugin to move.
    @type plugin: L{Plugin}
    @param new_view_name: New view name.
    @type new_view_name: string
    '''
    if not plugin or not isinstance(plugin, gtk.Widget): return
    old_view = plugin.get_parent()
    new_view = self._getViewOrNewView(new_view_name)
    if old_view is not new_view:
      old_view.remove(plugin)
      new_view.append_page(plugin)
      new_view.set_tab_detachable(plugin, True)
      new_view.set_tab_reorderable(plugin, True)

  def Menu(self, context_plugin, transient_window):
    '''
    Helps emulate a non-static inner class. These don't exist in python,
    I think.

    @param context_plugin: Subject plugin of this menu.
    @type context_plugin: L{Plugin}
    @param transient_window: Transient parent window. Used for keeping the
    new view dialog modal.
    @type transient_window: gtk.Window

    @return: An inner menu class.
    @rtype: L{ViewManager._Menu}
    '''
    return self._Menu(self, context_plugin, transient_window)

  class _Menu(gtk.Menu):
    '''
    Implements a popup menu for a plugin that will allow putting the plugin in
    a different view.

    @cvar RADIO_GROUP: Radio menu item's group id.
    @type RADIO_GROUP: integer

    @ivar view_manager: View manager to use as data model and controller.
    @type view_manager: L{ViewManager}
    '''
    RADIO_GROUP = 13
    def __init__(self, view_manager, context_plugin, transient_window):
      '''
      Initialize menu.

      @param view_manager: View manager to use as data model and controller.
      @type view_manager: L{ViewManager}
      @param context_plugin: Subject plugin of this menu.
      @type context_plugin: L{Plugin}
      @param transient_window: Transient parent window. Used for keeping the
      new view dialog modal.
      @type transient_window: gtk.Window
      '''
      gtk.Menu.__init__(self)
      self.view_manager = view_manager
      if isinstance(context_plugin, gtk.Widget):
        self._buildMenu(context_plugin, transient_window)

    def _buildMenu(self, context_plugin, transient_window):
      '''
      Build the menu according to the view managers model.

      @param context_plugin: Subject plugin of this menu.
      @type context_plugin: L{Plugin}
      @param transient_window: Transient parent window. Used for keeping the
      new view dialog modal.
      @type transient_window: gtk.Window
      '''
      menu_item = None
      for view in self.view_manager:
        menu_item = gtk.RadioMenuItem(label = _(view.view_name))
        menu_item.set_name(view.view_name)
        menu_item.connect('toggled', self._onItemToggled, view, context_plugin)
        menu_item.set_active(view == context_plugin.get_parent())
        self.append(menu_item)
        menu_item.show()
      menu_item = gtk.SeparatorMenuItem()
      self.append(menu_item)
      menu_item.show()
      menu_item = gtk.MenuItem(label="<i>" + _('_New view…') + "</i>")
      menu_item.get_child().set_use_markup(True)
      menu_item.connect('activate', self._onItemActivated,
                        context_plugin, transient_window)
      self.append(menu_item)
      menu_item.show()

    def _onItemToggled(self, menu_item, view, context_plugin):
      '''
      Callback for radio item toggles. Change the views accordingly.

      @param menu_item: Menu item that was toggled
      @type menu_item: gtk.RadioMenuItem
      @param view: View that was chosen.
      @type view: L{PluginView}
      @param context_plugin: Subject plugin of this menu.
      @type context_plugin: L{Plugin}
      '''
      self.view_manager.changeView(context_plugin, view.view_name)

    def _onItemActivated(self, menu_item, context_plugin, transient_window):
      '''
      Callback for "new view" menu item. Creates a dialog for
      entering a view name.

      @param menu_item: Menu item that was activated.
      @type menu_item: gtk.MenuItem
      @param context_plugin: Subject plugin of this menu.
      @type context_plugin: L{Plugin}
      @param transient_window: Transient parent window. Used for keeping the
      new view dialog modal.
      @type transient_window: gtk.Window
      '''
      new_view_dialog = \
          self._NewViewDialog(self.view_manager, transient_window)
      response_id = new_view_dialog.run()
      plugin_name = new_view_dialog.getEntryText()
      if response_id == gtk.ResponseType.OK and plugin_name:
        self.view_manager.changeView(context_plugin, plugin_name)
      new_view_dialog.destroy()

    class _NewViewDialog(gtk.Dialog):
      '''
      Small dialog that allows entry of a new view name.
      '''
      def __init__(self, view_manager, transient_window):
        '''


        @param view_manager: View manager to use as data model and controller.
        @type view_manager: L{ViewManager}
        @param transient_window: Transient parent window. Used for keeping the
        new view dialog modal.
        @type transient_window: gtk.Window
        '''
        self.view_manager = view_manager
        gtk.Dialog.__init__(self, _('New View…'), transient_window)
        ok_button = self.add_button(_('_OK'), gtk.ResponseType.OK)
        ok_button.set_image(gtk.Image.new_from_icon_name("gtk-apply", gtk.IconSize.BUTTON))
        close_button = self.add_button(_('_Close'), gtk.ResponseType.CLOSE)
        close_button.set_image(gtk.Image.new_from_icon_name("window-close", gtk.IconSize.BUTTON))
        self.set_default_response(gtk.ResponseType.OK)
        completion = gtk.EntryCompletion()
        complete_model = gtk.ListStore(str)
        for view in self.view_manager:
          complete_model.append([view.view_name])
        completion.set_model(complete_model)
        completion.set_text_column(0)
        self.entry = gtk.Entry()
        self.entry.set_completion(completion)
        self.entry.connect('activate', self._onEntryActivate)
        self.box = self.get_children()[0]
        self.box.add(self.entry)
        self.entry.show()

      def getEntryText(self):
        '''
        Get the contents of the entry widget.

        @return: Text in entry box.
        @rtype: string
        '''
        return self.entry.get_text()

      def _onEntryActivate(self, entry):
        '''
        Callback for activation of the entry box. Return an OK response.

        @param entry: Entry box that was activated.
        @type entry: gtk.Entry
        '''
        self.response(gtk.ResponseType.OK)

